"""Straight heater meander doped."""

from __future__ import annotations

from collections.abc import Iterable
from functools import partial

import gdsfactory as gf
from gdsfactory.component import Component, ComponentReference
from gdsfactory.components.vias.via import via
from gdsfactory.components.vias.via_stack import via_stack
from gdsfactory.cross_section import Section
from gdsfactory.typings import ComponentSpec, CrossSectionSpec, Floats, LayerSpecs, Port

_via_stack = partial(
    via_stack,
    size=(1.5, 1.5),
    layers=("M1", "M2"),
    vias=(
        partial(
            via,
            layer="VIAC",
            size=(0.1, 0.1),
            pitch=0.2,
            enclosure=0.1,
        ),
        partial(
            via,
            layer="VIA1",
            size=(0.1, 0.1),
            pitch=0.2,
            enclosure=0.1,
        ),
    ),
)


@gf.cell_with_module_name
def straight_heater_meander_doped(
    length: float = 300.0,
    spacing: float = 2.0,
    cross_section: CrossSectionSpec = "strip",
    heater_width: float = 1.5,
    extension_length: float = 15.0,
    layers_doping: LayerSpecs = ("P", "PP", "PPP"),
    radius: float = 5.0,
    via_stack: ComponentSpec | None = _via_stack,
    port_orientation1: float | None = None,
    port_orientation2: float | None = None,
    straight_widths: Floats = (0.8, 0.9, 0.8),
    taper_length: float = 10,
) -> Component:
    """Returns a meander based heater.

    based on SungWon Chung, Makoto Nakai, and Hossein Hashemi,
    Low-power thermo-optic silicon modulator for large-scale photonic integrated systems
    Opt. Express 27, 13430-13459 (2019)
    https://www.osapublishing.org/oe/abstract.cfm?URI=oe-27-9-13430

    Args:
        length: total length of the optical path.
        spacing: waveguide spacing (center to center).
        cross_section: for waveguide.
        heater_width: for heater.
        extension_length: of input and output optical ports.
        layers_doping: doping layers to be used for heater.
        radius: for the meander bends.
        via_stack: for the heater to via_stack metal.
        port_orientation1: in degrees. None adds all orientations.
        port_orientation2: in degrees. None adds all orientations.
        straight_widths: width of the straight sections.
        taper_length: from the cross_section.
    """
    from gdsfactory.pdk import get_layer

    rows = len(straight_widths)
    c = gf.Component()
    x = gf.get_cross_section(cross_section)
    layer = get_layer(x.layer)

    temp_component = Component()
    p1 = temp_component.add_port(
        name="p1",
        center=(0, 0),
        orientation=0,
        cross_section=x,
        layer=layer,
        width=x.width,
    )
    p2 = temp_component.add_port(
        name="p2",
        center=(0, spacing),
        orientation=0,
        cross_section=x,
        layer=layer,
        width=x.width,
    )

    dummy = gf.Component()
    route = gf.routing.route_single(
        dummy, p1, p2, radius=radius, cross_section=cross_section
    )
    cross_section2 = cross_section

    straight_length = gf.snap.snap_to_grid2x(
        (length - (rows - 1) * c.kcl.dbu * route.length) / rows,
    )
    ports: dict[str, Port] = {}

    if straight_length - 2 * taper_length <= 0:
        raise ValueError("straight_length - 2 * taper_length <= 0")

    # Straights
    for row, straight_width in enumerate(straight_widths):
        cross_section1 = gf.get_cross_section(cross_section, width=straight_width)
        straight = gf.c.straight(
            length=straight_length - 2 * taper_length, cross_section=cross_section1
        )

        taper = partial(
            gf.c.taper_cross_section_linear,
            cross_section1=cross_section1,
            cross_section2=cross_section2,
            length=taper_length,
        )

        straight_with_tapers = gf.c.extend_ports(straight, extension=taper)
        straight_ref = c << straight_with_tapers
        if row < len(straight_widths) // 2:
            straight_ref.y = row * spacing
        else:
            straight_ref.y = (row + 1) * spacing
        ports[f"o1_{row + 1}"] = straight_ref["o1"]
        ports[f"o2_{row + 1}"] = straight_ref["o2"]

    # Loopbacks
    for row in range(1, rows, 2):
        extra_length = 3 * (rows - row - 1) / 2 * radius
        extra_straight1 = c << gf.c.straight(
            length=extra_length, cross_section=cross_section
        )
        extra_straight1.connect("o1", ports[f"o1_{row + 1}"])
        extra_straight2 = c << gf.c.straight(
            length=extra_length, cross_section=cross_section
        )
        extra_straight2.connect("o1", ports[f"o1_{row + 2}"])

        route = gf.routing.route_single(
            c,
            extra_straight2["o2"],
            extra_straight1["o2"],
            radius=radius,
            cross_section=cross_section,
        )

        extra_length = 3 * (row - 1) / 2 * radius
        extra_straight1 = c << gf.c.straight(
            length=extra_length, cross_section=cross_section
        )
        extra_straight1.connect("o1", ports[f"o2_{row + 1}"])
        extra_straight2 = c << gf.c.straight(
            length=extra_length, cross_section=cross_section
        )
        extra_straight2.connect("o1", ports[f"o2_{row}"])

        route = gf.routing.route_single(
            c,
            extra_straight2.ports["o2"],
            extra_straight1.ports["o2"],
            radius=radius,
            cross_section=cross_section,
        )

    straight1 = c << gf.c.straight(length=extension_length, cross_section=cross_section)
    straight2 = c << gf.c.straight(length=extension_length, cross_section=cross_section)
    straight1.connect("o2", ports["o1_1"])
    straight2.connect("o1", ports[f"o2_{rows}"])

    c.add_port("o1", port=straight1.ports["o1"])
    c.add_port("o2", port=straight2.ports["o2"])

    heater: ComponentReference | None = None

    if layers_doping:
        sections: tuple[Section, ...] = ()
        for doping_layer in layers_doping:
            sections += (Section(layer=doping_layer, width=heater_width, offset=0),)
        heater_cross_section = partial(
            gf.cross_section.cross_section,
            width=heater_width,
            layer="WG",
            sections=sections,
            port_names=("e1", "e2"),
            port_types=("electrical", "electrical"),
        )

        heater = c << gf.c.straight(
            length=straight_length,
            cross_section=heater_cross_section,
        )
        heater.movey(spacing * (rows // 2))

    if layers_doping and via_stack and heater is not None:
        via = via_stacke = via_stackw = gf.get_component(via_stack)
        via_stack_west = c << via_stackw
        via_stack_east = c << via_stacke
        via_stack_west.connect(
            "e3", heater["e1"], allow_layer_mismatch=True, allow_width_mismatch=True
        )
        via_stack_east.connect(
            "e1", heater["e2"], allow_layer_mismatch=True, allow_width_mismatch=True
        )

        valid_orientations = {p.orientation for p in via.ports}
        ports1: Iterable[Port] = []
        ports2: Iterable[Port] = []
        if port_orientation1 is None:
            ports1 = via_stack_west.ports
        else:
            ports1 = via_stack_west.ports.filter(orientation=port_orientation1)

        if port_orientation2 is None:
            ports2 = via_stack_east.ports
        else:
            ports2 = via_stack_east.ports.filter(orientation=port_orientation2)

        if not ports1:
            raise ValueError(
                f"No ports for port_orientation1 {port_orientation1} in {valid_orientations}"
            )
        if not ports2:
            raise ValueError(
                f"No ports for port_orientation2 {port_orientation2} in {valid_orientations}"
            )

        c.add_ports(ports1, prefix="l_")
        c.add_ports(ports2, prefix="r_")

    # delete any straights with zero length
    for inst in list(c.insts):
        if inst.cell.settings.get("length") == 0.0:
            del c.insts[inst]
    c.flatten()
    return c



from __future__ import annotations

from typing import Any

import numpy as np
import numpy.typing as npt

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.config import ErrorType
from gdsfactory.functions import angles_deg, curvature, snap_angle
from gdsfactory.typings import Coordinate, Coordinates, CrossSectionSpec, Size


def bezier_curve(
    t: npt.NDArray[np.floating[Any]], control_points: Coordinates
) -> npt.NDArray[np.floating[Any]]:
    """Returns bezier coordinates.

    Args:
        t: 1D array of points varying between 0 and 1.
        control_points: for the bezier curve.
    """
    from scipy.special import binom

    xs = 0.0
    ys = 0.0
    n = len(control_points) - 1
    for k in range(n + 1):
        ank = binom(n, k) * (1 - t) ** (n - k) * t**k
        xs += ank * control_points[k][0]
        ys += ank * control_points[k][1]

    return np.column_stack([xs, ys])


@gf.cell_with_module_name
def bezier(
    control_points: Coordinates = ((0.0, 0.0), (5.0, 0.0), (5.0, 1.8), (10.0, 1.8)),
    npoints: int = 201,
    with_manhattan_facing_angles: bool = True,
    start_angle: int | None = None,
    end_angle: int | None = None,
    cross_section: CrossSectionSpec = "strip",
    bend_radius_error_type: ErrorType | None = None,
    allow_min_radius_violation: bool = False,
    width: float | None = None,
) -> Component:
    """Returns Bezier bend.

    Args:
        control_points: list of points.
        npoints: number of points varying between 0 and 1.
        with_manhattan_facing_angles: bool.
        start_angle: optional start angle in deg.
        end_angle: optional end angle in deg.
        cross_section: spec.
        bend_radius_error_type: error type.
        allow_min_radius_violation: bool.
        width: width to use. Defaults to cross_section.width.
    """
    if width:
        xs = gf.get_cross_section(cross_section, width=width)
    else:
        xs = gf.get_cross_section(cross_section)

    t = np.linspace(0, 1, npoints)
    path_points = bezier_curve(t, control_points)
    path = gf.Path(path_points)

    if with_manhattan_facing_angles:
        path.start_angle = start_angle or snap_angle(path.start_angle)
        path.end_angle = end_angle or snap_angle(path.end_angle)

    c = path.extrude(xs)
    curv = curvature(path_points, t)
    length = path.length()
    if max(np.abs(curv)) == 0:
        min_bend_radius = np.inf
    else:
        min_bend_radius = float(gf.snap.snap_to_grid(float(1 / np.max(np.abs(curv)))))

    c.info["length"] = length
    c.info["min_bend_radius"] = min_bend_radius
    c.info["start_angle"] = float(path.start_angle)
    c.info["end_angle"] = float(path.end_angle)
    c.add_route_info(
        cross_section=xs,
        length=c.info["length"],
        n_bend_s=1,
        min_bend_radius=min_bend_radius,
    )

    if not allow_min_radius_violation:
        xs.validate_radius(min_bend_radius, bend_radius_error_type)

    xs.add_bbox(c)
    return c


def find_min_curv_bezier_control_points(
    start_point: Coordinate,
    end_point: Coordinate,
    start_angle: float,
    end_angle: float,
    npoints: int = 201,
    alpha: float = 0.05,
    nb_pts: int = 2,
) -> Coordinates:
    """Returns bezier control points that minimize curvature.

    Args:
        start_point: start point.
        end_point: end point.
        start_angle: start angle in deg.
        end_angle: end angle in deg.
        npoints: number of points varying between 0 and 1.
        alpha: weight for angle mismatch.
        nb_pts: number of control points.
    """
    from scipy.optimize import minimize

    t = np.linspace(0, 1, npoints)

    def array_1d_to_cpts(a: npt.NDArray[np.float64]) -> list[tuple[float, float]]:
        xs = a[::2]
        ys = a[1::2]
        return list(zip(xs, ys, strict=False))

    def objective_func(p: npt.NDArray[np.float64]) -> float:
        """Minimize  max curvaturea and negligible start angle and end angle mismatch."""
        ps = array_1d_to_cpts(p)
        control_points = [start_point] + ps + [end_point]
        path_points = bezier_curve(t, control_points)

        max_curv = max(np.abs(curvature(path_points, t)))

        angles = angles_deg(path_points)
        dstart_angle = abs(angles[0] - start_angle)
        dend_angle = abs(angles[-2] - end_angle)
        angle_mismatch = dstart_angle + dend_angle
        return float(angle_mismatch * alpha + max_curv)

    x0, y0 = start_point[0], start_point[1]
    xn, yn = end_point[0], end_point[1]

    initial_guess: list[float] = []
    for i in range(nb_pts):
        x = (i + 1) * (x0 + xn) / nb_pts
        y = (i + 1) * (y0 + yn) / nb_pts
        initial_guess += [x, y]

    # initial_guess = [(x0 + xn) / 2, y0, (x0 + xn) / 2, yn]
    res = minimize(objective_func, initial_guess, method="Nelder-Mead")
    p = res.x
    points = [start_point] + array_1d_to_cpts(p) + [end_point]
    return tuple(points)


@gf.cell_with_module_name
def bend_s(
    size: Size = (11.0, 1.8),
    npoints: int = 99,
    cross_section: CrossSectionSpec = "strip",
    allow_min_radius_violation: bool = False,
    width: float | None = None,
) -> Component:
    """Return S bend with bezier curve.

    stores min_bend_radius property in self.info['min_bend_radius']
    min_bend_radius depends on height and length

    Args:
        size: in x and y direction.
        npoints: number of points.
        cross_section: spec.
        allow_min_radius_violation: bool.
        width: width to use. Defaults to cross_section.width.

    """
    dx, dy = size

    if dy == 0:
        return gf.components.straight(
            length=dx, cross_section=cross_section, width=width
        )

    return bezier(
        control_points=((0, 0), (dx / 2, 0), (dx / 2, dy), (dx, dy)),
        npoints=npoints,
        cross_section=cross_section,
        allow_min_radius_violation=allow_min_radius_violation,
        width=width,
    )


def _get_arc_sbend_angle_middle_length_from_jog(
    jog: float, radius: float
) -> tuple[float, float]:
    """Compute the Euler bend angle and middle straight length for an S-bend."""
    if jog < 2 * radius:
        angle = np.rad2deg(2 * np.arcsin(np.sqrt(jog / (4 * radius))))
        middle_length = 0.0
    else:
        angle = 90.0
        middle_length = jog - 2 * radius
    return (angle, middle_length)


def _get_euler_sbend_angle_middle_length_from_jog(
    jog: float, radius: float
) -> tuple[float, float]:
    """Compute the Euler bend angle (in degrees) and middle straight length for an S-bend.

    using SciPy to numerically solve for the bend angle required to achieve half the jog.

    The vertical displacement for an Euler bend of angle θ (in radians) is given by:
      displacement(θ) = radius * sqrt(pi * θ) * S( sqrt(2θ/pi) )
    where S() is the Fresnel sine integral.

    The S-bend consists of two symmetric Euler bends. If the jog is less than twice the displacement
    of a full 90° Euler bend, the angle is computed such that one Euler bend gives a displacement of jog/2.
    Otherwise, a full 90° Euler bend is used and the extra required offset is added as a straight section.

    Args:
        jog: The vertical displacement of the S-bend.
        radius: The radius of the Euler bend.

    Returns:
      tuple: (angle_deg, middle_length) where:
          - angle_deg is the Euler bend angle in degrees.
          - middle_length is the length of the straight segment between the Euler bends.
    """
    from scipy import optimize

    def euler_displacement(theta: float) -> float:
        curve = gf.path.euler(radius=radius, angle=theta, use_eff=False, p=1)
        return curve.ysize

    dy_full = euler_displacement(theta=90)

    if jog < 2 * dy_full:
        # Define the objective function: squared error between computed displacement and jog
        def objective(theta: float) -> float:
            return (euler_displacement(theta) - jog) ** 2

        result = optimize.minimize_scalar(objective, bounds=(1, 90), method="bounded")
        angle_deg = result.x
        middle_length = 0.0
    else:
        angle_deg = 90.0
        middle_length = jog - 2 * dy_full

    return angle_deg, middle_length


@gf.cell_with_module_name
def bend_s_offset(
    offset: float = 40.0,
    radius: float | None = 10.0,
    cross_section: CrossSectionSpec = "strip",
    width: float | None = None,
    with_euler: bool = True,
) -> gf.Component:
    """Return S bend made of two euler bends with a straight section.

    stores min_bend_radius property in self.info['min_bend_radius']
    min_bend_radius depends on height and length

    Args:
        offset: in um.
        radius: in um. if None, uses cross_section_radius.
        cross_section: spec.
        width: width to use. Defaults to cross_section.width.
        with_euler: use euler bend instead of arc bend.
    """
    if width:
        xs = gf.get_cross_section(cross_section, width=width)
    else:
        xs = gf.get_cross_section(cross_section)

    radius = radius or xs.radius
    assert radius is not None, "radius cannot be None"

    xs.validate_radius(radius)
    if with_euler:
        angle, middle_length = _get_euler_sbend_angle_middle_length_from_jog(
            jog=offset / 2, radius=radius
        )
        path = gf.path.euler(radius=radius, angle=+angle, p=1, use_eff=False)
        path += gf.path.straight(length=middle_length)
        path += gf.path.euler(radius=radius, angle=-angle, p=1, use_eff=False)
    else:
        angle, middle_length = _get_arc_sbend_angle_middle_length_from_jog(
            jog=offset,
            radius=radius,
        )
        path = gf.path.arc(radius=radius, angle=+angle)
        path += gf.path.straight(length=middle_length)
        path += gf.path.arc(radius=radius, angle=-angle)

    return gf.path.extrude(path, cross_section=xs)


def get_min_sbend_size(
    size: tuple[float | None, float | None] = (None, 10.0),
    cross_section: CrossSectionSpec = "strip",
    num_points: int = 100,
) -> float:
    """Returns the minimum sbend size to comply with bend radius requirements.

    Args:
        size: in x and y direction. One of them is None, which is the size we need to figure out.
        cross_section: spec.
        num_points: number of points to iterate over between max_size and 0.1 * max_size.
    """
    size_list = list(size)
    cross_section_f = gf.get_cross_section(cross_section)

    if size_list[0] is None:
        ind = 0
        known_s = size_list[1]
    elif size_list[1] is None:
        ind = 1
        known_s = size_list[0]
    else:
        raise ValueError("One of the two elements in size has to be None")

    min_radius = cross_section_f.radius

    if min_radius is None:
        raise ValueError("The min radius for the specified layer is not known!")

    min_size = np.inf

    assert known_s is not None

    # Guess sizes, iterate over them until we cannot achieve the min radius
    # the max size corresponds to an ellipsoid
    max_size = float(np.sqrt(np.abs(min_radius * known_s)) * 2.5)
    sizes = np.linspace(max_size, 0.1 * max_size, num_points)

    for s in sizes:
        sz = size_list
        sz[ind] = s
        dx, dy = size_list
        assert dx is not None and dy is not None
        control_points = ((0, 0), (dx / 2, 0), (dx / 2, dy), (dx, dy))
        npoints = 201
        t = np.linspace(0, 1, npoints)
        path_points = bezier_curve(t, control_points)
        curv = curvature(path_points, t)
        min_bend_radius = 1 / max(np.abs(curv))
        if min_bend_radius < min_radius:
            min_size = s
            break

    return min_size



from gdsfactory.components.pcms.cavity import (
    cavity,
)
from gdsfactory.components.pcms.cdsem_all import (
    cdsem_all,
)
from gdsfactory.components.pcms.cdsem_bend180 import (
    cdsem_bend180,
)
from gdsfactory.components.pcms.cdsem_coupler import (
    cdsem_coupler,
)
from gdsfactory.components.pcms.cdsem_straight import (
    cdsem_straight,
)
from gdsfactory.components.pcms.cdsem_straight_density import (
    cdsem_straight_density,
    gaps,
    widths,
)
from gdsfactory.components.pcms.cutback_2x2 import (
    cutback_2x2,
)
from gdsfactory.components.pcms.cutback_bend import (
    cutback_bend,
    cutback_bend90,
    cutback_bend90circular,
    cutback_bend180,
    cutback_bend180circular,
    staircase,
)
from gdsfactory.components.pcms.cutback_component import (
    cutback_component,
    cutback_component_mirror,
)
from gdsfactory.components.pcms.cutback_loss import (
    cutback_loss,
    cutback_loss_bend90,
    cutback_loss_bend180,
    cutback_loss_mmi1x2,
    cutback_loss_spirals,
)
from gdsfactory.components.pcms.cutback_splitter import (
    cutback_splitter,
)
from gdsfactory.components.pcms.greek_cross import (
    greek_cross,
    greek_cross_with_pads,
)
from gdsfactory.components.pcms.litho_calipers import (
    litho_calipers,
)
from gdsfactory.components.pcms.litho_ruler import (
    litho_ruler,
)
from gdsfactory.components.pcms.litho_steps import (
    litho_steps,
)
from gdsfactory.components.pcms.resistance_meander import (
    resistance_meander,
)
from gdsfactory.components.pcms.resistance_sheet import (
    resistance_sheet,
)
from gdsfactory.components.pcms.ruler import (
    ruler,
)
from gdsfactory.components.pcms.verniers import (
    verniers,
)
from gdsfactory.components.pcms.version_stamp import (
    pixel,
    qrcode,
    version_stamp,
)

__all__ = [
    "cavity",
    "cdsem_all",
    "cdsem_bend180",
    "cdsem_coupler",
    "cdsem_straight",
    "cdsem_straight_density",
    "cutback_2x2",
    "cutback_bend",
    "cutback_bend90",
    "cutback_bend90circular",
    "cutback_bend180",
    "cutback_bend180circular",
    "cutback_component",
    "cutback_component_mirror",
    "cutback_loss",
    "cutback_loss_bend90",
    "cutback_loss_bend180",
    "cutback_loss_mmi1x2",
    "cutback_loss_spirals",
    "cutback_splitter",
    "gaps",
    "greek_cross",
    "greek_cross_with_pads",
    "litho_calipers",
    "litho_ruler",
    "litho_steps",
    "pixel",
    "qrcode",
    "resistance_meander",
    "resistance_sheet",
    "ruler",
    "staircase",
    "verniers",
    "version_stamp",
    "widths",
]



from __future__ import annotations

from collections.abc import Callable
from typing import Any

import numpy as np
import numpy.typing as npt

import gdsfactory as gf
from gdsfactory.path import transition_adiabatic
from gdsfactory.typings import CrossSectionSpec


def neff_TE1550SOI_220nm(w: float) -> float:
    """Returns the effective index of the fundamental TE mode for a 220nm-thick core with 3.45 index, fully clad with 1.44 index.

    Args:
        w: width in um.

    Returns:
        effective index.
    """
    adiabatic_polyfit_TE1550SOI_220nm = np.array(
        [
            1.02478963e-09,
            -8.65556534e-08,
            3.32415694e-06,
            -7.68408985e-05,
            1.19282177e-03,
            -1.31366332e-02,
            1.05721429e-01,
            -6.31057637e-01,
            2.80689677e00,
            -9.26867694e00,
            2.24535191e01,
            -3.90664800e01,
            4.71899278e01,
            -3.74726005e01,
            1.77381560e01,
            -1.12666286e00,
        ]
    )
    return float(np.poly1d(adiabatic_polyfit_TE1550SOI_220nm)(w).item())


@gf.cell_with_module_name
def taper_adiabatic(
    width1: float = 0.5,
    width2: float = 5.0,
    length: float = 0,
    neff_w: Callable[[float], float] = neff_TE1550SOI_220nm,
    alpha: float = 1,
    wavelength: float = 1.55,
    npoints: int = 200,
    cross_section: CrossSectionSpec = "strip",
    max_length: float = 200,
) -> gf.Component:
    """Returns a straight adiabatic_taper from an effective index callable.

    Args:
        width1: initial width.
        width2: final width.
        length: 0 uses the optimized length, and otherwise the optimal shape is compressed/stretched to the specified length.
        neff_w: a callable that returns the effective index as a function of width
                - By default, will use a compact model of neff(y) for fundamental 1550 nm TE mode of 220nm-thick core with 3.45 index, fully clad with 1.44 index. Many coefficients are needed to capture the behaviour.
        alpha: parameter that scales the rate of width change.
                - closer to 0 means longer and more adiabatic;
                - 1 is the intuitive limit beyond which higher order modes are excited;
                - [2] reports good performance up to 1.4 for fundamental TE in SOI (for multiple core thicknesses)
        wavelength: wavelength in um.
        npoints: number of points for sampling.
        cross_section: cross_section specification.
        max_length: maximum length for the taper.

    References:
        [1] Burns, W. K., et al. "Optical waveguide parabolic coupling horns." Appl. Phys. Lett., vol. 30, no. 1, 1 Jan. 1977, pp. 28-30, doi:10.1063/1.89199.
        [2] Fu, Yunfei, et al. "Efficient adiabatic silicon-on-insulator waveguide taper." Photonics Res., vol. 2, no. 3, 1 June 2014, pp. A41-A44, doi:10.1364/PRJ.2.000A41.
        npoints: number of points for sampling
    """
    xs = gf.get_cross_section(cross_section)
    layer = xs.layer
    assert layer is not None

    # Obtain optimal curve
    x_opt, w_opt = transition_adiabatic(
        width1,
        width2,
        neff_w=neff_w,
        wavelength=wavelength,
        alpha=alpha,
        max_length=max_length,
    )

    # Resample the points
    from scipy import interpolate

    w_opt_interp = interpolate.interp1d(x_opt, w_opt)

    if not length:
        length = x_opt[-1]
    x = np.linspace(0, length, npoints)
    w: npt.NDArray[np.floating[Any]] = w_opt_interp(x)

    assert isinstance(w, np.ndarray)

    # Stretch/compress x
    x_array: npt.NDArray[np.float64] = np.linspace(0, length, npoints) * (
        1 + length - x_opt[-1]
    )
    assert isinstance(x_array, np.ndarray)
    y_array = w / 2

    c = gf.Component()
    c.add_polygon(
        [(float(x), float(y)) for x, y in zip(x_array, y_array, strict=False)]
        + [(float(x), float(y)) for x, y in zip(x_array, -y_array, strict=False)][::-1],
        layer=layer,
    )

    # Define ports
    c.add_port(
        name="o1",
        center=(0, 0),
        width=width1,
        orientation=180,
        cross_section=cross_section,
        layer=layer,
    )
    c.add_port(
        name="o2",
        center=(length, 0),
        width=width2,
        orientation=0,
        cross_section=cross_section,
        layer=layer,
    )
    xs.add_bbox(c)
    return c



"""Serialize component settings into YAML or strings."""

from __future__ import annotations

import functools
import hashlib
import inspect
import pathlib
import re
from collections.abc import KeysView
from keyword import iskeyword
from typing import Any, overload

import attrs
import numpy as np
import orjson
import pydantic
import toolz
import toolz.functoolz
from aenum import Enum

DEFAULT_SERIALIZATION_MAX_DIGITS = 3
"""By default, the maximum number of digits retained when serializing float-like arrays"""


def get_string(value: Any) -> str:
    try:
        s = orjson.dumps(
            value, option=orjson.OPT_SERIALIZE_NUMPY, default=clean_value_name
        ).decode()
    except TypeError as e:
        print(f"Error serializing {value!r}")
        raise e
    return str(s)


def clean_dict(dictionary: dict[str, Any]) -> dict[str, Any]:
    return {k: clean_value_json(v) for k, v in dictionary.items()}


def complex_encoder(
    obj: complex | np.complexfloating, digits: int = DEFAULT_SERIALIZATION_MAX_DIGITS
) -> dict[str, Any]:
    real_part = np.round(obj.real, digits)
    imag_part = np.round(obj.imag, digits)
    return {"real": real_part, "imag": imag_part}


@overload
def clean_value_json(
    value: dict[str, Any],
    include_module: bool = True,
    serialize_function_as_dict: bool = True,
) -> dict[str, Any]: ...


@overload
def clean_value_json(
    value: str,
    include_module: bool = True,
    serialize_function_as_dict: bool = True,
) -> str: ...


@overload
def clean_value_json(
    value: Any,
    include_module: bool = True,
    serialize_function_as_dict: bool = True,
) -> str | int | float | dict[str, Any] | list[Any] | bool | Any | None: ...


def clean_value_json(
    value: Any, include_module: bool = True, serialize_function_as_dict: bool = True
) -> str | int | float | dict[str, Any] | list[Any] | bool | Any | None:
    """Return JSON serializable object.

    Args:
        value: object to serialize.
        include_module: include module in serialization.
        serialize_function_as_dict: serialize function as dict. False serializes as string.
    """
    from gdsfactory.path import Path

    if isinstance(value, pydantic.BaseModel):
        return clean_dict(value.model_dump(exclude_none=True))

    elif hasattr(value, "get_component_spec"):
        return value.get_component_spec()

    elif isinstance(value, bool):
        return value

    elif isinstance(value, Enum):
        return str(value)

    elif isinstance(value, np.integer | int):
        return int(value)

    elif isinstance(value, float | np.floating):
        if value == round(value):
            return int(value)
        return float(np.round(value, DEFAULT_SERIALIZATION_MAX_DIGITS))

    elif isinstance(value, complex | np.complexfloating):
        return complex_encoder(value)

    elif isinstance(value, np.ndarray):
        value = np.round(value, DEFAULT_SERIALIZATION_MAX_DIGITS)
        return orjson.loads(orjson.dumps(value, option=orjson.OPT_SERIALIZE_NUMPY))

    elif callable(value) and isinstance(value, functools.partial):
        return clean_value_partial(
            value=value,
            include_module=include_module,
            serialize_function_as_dict=serialize_function_as_dict,
        )
    elif hasattr(value, "to_dict"):
        return clean_dict(value.to_dict())

    elif callable(value) and isinstance(value, toolz.functoolz.Compose):
        return [clean_value_json(value.first)] + [
            clean_value_json(func) for func in value.funcs
        ]

    elif callable(value) and hasattr(value, "__name__"):
        if serialize_function_as_dict:
            return (
                {"function": value.__name__, "module": value.__module__}
                if include_module
                else {"function": value.__name__}
            )
        else:
            return value.__name__

    elif isinstance(value, Path):
        return value.hash_geometry()

    elif isinstance(value, pathlib.Path):
        return value.stem

    elif isinstance(value, dict):
        return clean_dict(value.copy())

    elif isinstance(value, list | tuple | set | KeysView):
        return tuple([clean_value_json(i) for i in value])

    elif attrs.has(type(value)):
        return attrs.asdict(value)

    else:
        try:
            value_json = orjson.dumps(
                value, option=orjson.OPT_SERIALIZE_NUMPY, default=clean_value_json
            )
            return orjson.loads(value_json)
        except TypeError as e:
            print(f"Error serializing {value!r}")
            raise e


def clean_value_partial(
    value: functools.partial[Any],
    include_module: bool = True,
    serialize_function_as_dict: bool = True,
) -> str | Any | dict[str, str | Any | dict[str, Any]]:
    sig = inspect.signature(value.func)
    args_as_kwargs = dict(zip(sig.parameters.keys(), value.args, strict=False))
    args_as_kwargs |= value.keywords
    args_as_kwargs = clean_dict(args_as_kwargs)

    func = value.func
    while hasattr(func, "func"):
        func = func.func
    assert hasattr(func, "__name__")
    v = {
        "function": func.__name__,
        "settings": args_as_kwargs,
    }
    if include_module:
        assert hasattr(func, "__module__")
        v.update(module=func.__module__)
    if not serialize_function_as_dict:
        return func.__name__
    return v


def clean_value_name(value: Any) -> str:
    """Returns a valid Python variable name representation of an object."""
    # Convert the value to a string and replace spaces with underscores
    cleaned = str(clean_value_json(value)).replace(" ", "_")

    # Remove invalid characters (only allow letters, numbers, and underscores)
    cleaned = re.sub(r"[^a-zA-Z0-9_]", "", cleaned)

    # Ensure the name starts with a letter or underscore
    if not cleaned or not cleaned[0].isalpha():
        cleaned = f"var_{cleaned}"

    # Avoid reserved Python keywords
    if iskeyword(cleaned):
        cleaned = f"{cleaned}_var"

    return cleaned


def get_hash(value: Any) -> str:
    return hashlib.md5((clean_value_name(value)).encode()).hexdigest()[:8]



from gdsfactory.components.mmis.mmi import (
    mmi,
)
from gdsfactory.components.mmis.mmi1x2 import (
    mmi1x2,
)
from gdsfactory.components.mmis.mmi1x2_with_sbend import (
    mmi1x2_with_sbend,
    mmi_widths,
)
from gdsfactory.components.mmis.mmi2x2 import (
    mmi2x2,
)
from gdsfactory.components.mmis.mmi2x2_with_sbend import (
    mmi2x2_with_sbend,
)
from gdsfactory.components.mmis.mmi_90degree_hybrid import (
    mmi_90degree_hybrid,
)
from gdsfactory.components.mmis.mmi_tapered import (
    mmi_tapered,
)

__all__ = [
    "mmi",
    "mmi1x2",
    "mmi1x2_with_sbend",
    "mmi2x2",
    "mmi2x2_with_sbend",
    "mmi_90degree_hybrid",
    "mmi_tapered",
    "mmi_widths",
]



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.typings import CrossSectionSpec, Size


@gf.cell
def big_device(
    size: Size = (400.0, 400.0),
    nports: int = 16,
    spacing: float = 15.0,
    port_type: str = "electrical",
    cross_section: CrossSectionSpec = "metal_routing",
) -> Component:
    """Big component with N ports on each side.

    Args:
        size: x, y.
        nports: number of ports.
        spacing: in um.
        port_type: optical, electrical, rf, etc.
        cross_section: spec.
    """
    component = gf.Component()

    w, h = size
    dx = w / 2
    dy = h / 2
    n = nports

    xs = gf.get_cross_section(cross_section)
    layer = xs.layer
    assert layer is not None
    width = xs.width
    assert isinstance(width, float)

    points = [(dx, dy), (dx, -dy), (-dx, -dy), (-dx, dy)]
    component.add_polygon(points, layer=layer)

    for i in range(n):
        component.add_port(
            name=f"W{i}",
            center=(-dx, (i - n / 2) * spacing),
            orientation=180,
            port_type=port_type,
            cross_section=xs,
            layer=gf.get_layer(layer),
            width=width,
        )

    for i in range(n):
        component.add_port(
            name=f"E{i}",
            center=(dx, (i - n / 2) * spacing),
            orientation=0,
            port_type=port_type,
            cross_section=xs,
            layer=gf.get_layer(layer),
            width=width,
        )

    for i in range(n):
        component.add_port(
            name=f"N{i}",
            center=((i - n / 2) * spacing, dy),
            orientation=90,
            port_type=port_type,
            cross_section=xs,
            layer=gf.get_layer(layer),
            width=width,
        )

    for i in range(n):
        component.add_port(
            name=f"S{i}",
            center=((i - n / 2) * spacing, -dy),
            orientation=-90,
            port_type=port_type,
            cross_section=xs,
            layer=gf.get_layer(layer),
            width=width,
        )

    component.auto_rename_ports()
    return component


if __name__ == "__main__":
    import gdsfactory as gf

    c = big_device()
    c = gf.routing.add_pads_top(c, fanout_length=None)
    c.show()



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.name import clean_name
from gdsfactory.snap import snap_to_grid as snap
from gdsfactory.typings import Layer

ignore = (
    "cross_section",
    "decorator",
    "cross_section1",
    "cross_section2",
    "contact",
    "pad",
)
prefix_to_type_default = {
    "opt": "OPTICALPORT",
    "pad": "ELECTRICALPORT",
    "vertical_dc": "ELECTRICALPORT",
    "optical": "OPTICALPORT",
    "loopback": "OPTICALPORT",
}


def add_label_ehva(
    component: gf.Component,
    die: str = "demo",
    prefix_to_type: dict[str, str] = prefix_to_type_default,
    layer: Layer = (66, 0),
    metadata_ignore: list[str] | None = None,
    metadata_include_parent: list[str] | None = None,
    metadata_include_child: list[str] | None = None,
) -> gf.Component:
    """Returns Component with measurement labels.

    Args:
        component: to add labels to.
        die: string.
        prefix_to_type: dict of port prefix to port type.
        layer: text label layer.
        metadata_ignore: list of settings keys to ignore. Works with flatdict setting:subsetting.
        metadata_include_parent: includes parent metadata. Works with flatdict setting:subsetting.
        metadata_include_child: includes child metadata.

    """
    metadata_ignore = metadata_ignore or []
    metadata_include_parent = metadata_include_parent or []
    metadata_include_child = metadata_include_child or []

    text = f"""DIE NAME:{die}
CIRCUIT NAME:{component.name}
"""
    info = []

    metadata = component.settings
    if metadata:
        info += [
            f"CIRCUITINFO NAME: {k}, VALUE: {v}"
            for k, v in metadata
            if k not in metadata_ignore and isinstance(v, int | float | str)
        ]

    metadata = component.settings
    info += [
        f"CIRCUITINFO NAME: {clean_name(k)}, VALUE: {metadata.get(k)}"
        for k in metadata_include_parent
        if metadata.get(k)
    ]

    metadata = component.settings
    info += [
        f"CIRCUITINFO NAME: {k}, VALUE: {metadata.get(k)}"
        for k in metadata_include_child
        if metadata.get(k)
    ]

    text += "\n".join(info)
    text += "\n"

    info = []
    if component.ports:
        for prefix, port_type_ehva in prefix_to_type.items():
            info += [
                f"{port_type_ehva} NAME: {port.name} TYPE: {port_type_ehva}, "
                f"POSITION RELATIVE:({snap(port.x)}, {snap(port.y)}),"
                f" ORIENTATION: {port.orientation}"
                for port in component.get_ports_list(prefix=prefix)
            ]
    text += "\n".join(info)

    component.add_label(text=text, position=(0, 0), layer=layer)
    return component



from __future__ import annotations

import warnings
from collections.abc import Callable, Sequence
from functools import partial
from typing import TYPE_CHECKING, Any, Literal, TypeAlias

import kfactory as kf
import numpy as np
import numpy.typing as npt
from numpy import cos, float64, sin

import gdsfactory as gf

if TYPE_CHECKING:
    from gdsfactory.component import Component, ComponentReference
    from gdsfactory.typings import LayerSpec, LayerSpecs

RAD2DEG = 180.0 / np.pi
DEG2RAD = 1 / RAD2DEG


def move_port_to_zero(
    component: Component, port_name: str = "o1", mirror: bool = False
) -> gf.Component:
    """Return a container that contains a reference to the original component.

    The new component has port_name in (0, 0).

    Args:
        component: to move the port to (0, 0).
        port_name: to move to (0, 0).
        mirror: if True, mirrors the component.
    """
    port_names = [p.name for p in component.ports]
    if port_name not in port_names:
        raise ValueError(f"port_name = {port_name!r} not in {port_names}")

    c = gf.Component()
    ref = c << component
    if mirror:
        ref.dmirror()

    movement = np.array(ref.ports[port_name].center)
    ref.move(tuple(-movement))
    c.add_ports(ref.ports)
    c.copy_child_info(component)
    return c


def get_layers(component: Component) -> list[tuple[int, int]]:
    """Returns the layers of a component.

    Args:
        component: to get the layers from.
    """
    return [
        (info.layer, info.datatype)
        for info in component.kcl.layer_infos()
        if not component.bbox(component.kcl.layer(info)).empty()
    ]


def extract(
    component: Component,
    layers: LayerSpecs,
    recursive: bool = True,
) -> Component:
    """Extracts a list of layers and adds them to a new Component.

    Args:
        component: to extract the layers from.
        layers: list of layers to extract.
        recursive: if True, extracts the shapes recursively.
    """
    from gdsfactory.pdk import get_layer_tuple

    c = gf.Component()

    layer_tuples = [get_layer_tuple(layer) for layer in layers]
    component_layers = get_layers(component)

    for layer_tuple in layer_tuples:
        if layer_tuple not in component_layers:
            warnings.warn(
                f"Layer {layer_tuple} not found in component {component.name!r} layers. {component_layers}",
                stacklevel=3,
            )

    for layer_tuple in component_layers:
        if layer_tuple in layer_tuples:
            layer_index = c.kcl.layer(*layer_tuple)
            if recursive:
                c.shapes(layer_index).insert(component.begin_shapes_rec(layer_index))
            else:
                c.shapes(layer_index).insert(component.shapes(layer_index))

    return c


def move_to_center(component: Component, dx: float = 0, dy: float = 0) -> gf.Component:
    """Moves the component to the center of the bounding box."""
    c = component
    c.transform(gf.kdb.DTrans(-c.dbbox().center().x + dx, -c.dbbox().center().y + dy))
    return c


def move_port(
    component: Component, port_name: str, dx: float = 0, dy: float = 0
) -> gf.Component:
    """Moves the component port to a specific location.

    Warning: This function modifies the component in-place.

    Args:
        component: to move the port.
        port_name: to move.
        dx: to move the port.
        dy: to move the port.
    """
    c = component
    c.transform(gf.kdb.DTrans(-c.ports[port_name].x + dx, -c.ports[port_name].y + dy))
    return c


GetPolygonsResult: TypeAlias = "dict[LayerSpec, list[kf.kdb.Polygon]]"


def get_polygons(
    component_or_instance: Component | ComponentReference,
    merge: bool = False,
    by: Literal["index", "name", "tuple"] = "index",
    layers: LayerSpecs | None = None,
    smooth: float | None = None,
) -> GetPolygonsResult:
    """Returns a dict of Polygons per layer.

    Args:
        component_or_instance: to extract the polygons.
        merge: if True, merges the polygons.
        by: the format of the resulting keys in the dictionary ('index', 'name', 'tuple').
        layers: list of layer specs to extract the polygons from. If None, extracts all layers.
        smooth: if True, smooths the polygons.
    """
    from gdsfactory.pdk import get_layer, get_layer_name, get_layer_tuple

    if by == "index":
        get_key: Callable[[LayerSpec], LayerSpec] = get_layer
    elif by == "name":
        get_key = get_layer_name
    elif by == "tuple":
        get_key = get_layer_tuple
    else:
        raise ValueError("argument 'by' should be 'index' | 'name' | 'tuple'")

    polygons: GetPolygonsResult = {}

    c = component_or_instance
    if layers is None:
        layers = [
            (info.layer, info.datatype)
            for info in c.kcl.layer_infos()
            if not c.bbox(c.kcl.layer(info)).empty()
        ]

    layer_indexes = [get_layer(layer) for layer in layers]

    for layer_index in layer_indexes:
        layer_key = get_key(layer_index)
        if isinstance(component_or_instance, gf.Component):
            r = gf.kdb.Region(c.begin_shapes_rec(layer_index))
        else:
            r = kf.kdb.Region(c.cell.begin_shapes_rec(layer_index)).transformed(
                c.cplx_trans
            )
        if layer_key not in polygons:
            polygons[layer_key] = []
        if smooth:
            r.smooth(round(smooth / c.kcl.dbu))
        if merge:
            r.merge()
        for p in r.each():
            polygons[layer_key].append(p)
    return polygons


def get_polygons_points(
    component_or_instance: Component | ComponentReference,
    merge: bool = False,
    scale: float | None = None,
    by: Literal["index", "name", "tuple"] = "index",
    layers: LayerSpecs | None = None,
) -> dict[int | str | tuple[int, int], list[npt.NDArray[np.floating[Any]]]]:
    """Returns a dict with list of points per layer.

    Args:
        component_or_instance: to extract the polygons.
        merge: if True, merges the polygons.
        scale: if True, scales the points.
        by: the format of the resulting keys in the dictionary ('index', 'name', 'tuple').
        layers: list of layer specs to extract the polygons from. If None, extracts all layers.
    """
    polygons_dict = get_polygons(
        component_or_instance=component_or_instance, merge=merge, by=by, layers=layers
    )
    polygons_points: dict[
        tuple[int, int] | str | int, list[npt.NDArray[np.floating[Any]]]
    ] = {}
    for layer, polygons in polygons_dict.items():
        all_points: list[npt.NDArray[np.floating[Any]]] = []
        for polygon in polygons:
            if scale:
                points = np.array(
                    [
                        (point.x * scale, point.y * scale)
                        for point in polygon.to_simple_polygon()
                        .to_dtype(component_or_instance.kcl.dbu)
                        .each_point()
                    ]
                )
            else:
                points = np.array(
                    [
                        (point.x, point.y)
                        for point in polygon.to_simple_polygon()
                        .to_dtype(component_or_instance.kcl.dbu)
                        .each_point()
                    ]
                )
            all_points.append(points)
        polygons_points[layer] = all_points
    return polygons_points


def get_point_inside(
    component_or_instance: Component | ComponentReference, layer: LayerSpec
) -> npt.NDArray[np.floating[Any]]:
    """Returns a point inside the component or instance.

    Args:
        component_or_instance: to find a point inside.
        layer: to find a point inside.
    """
    layer = gf.get_layer(layer)
    return np.array(
        get_polygons_points(component_or_instance, layers=[layer])[layer][0][0]
    )


def sign_shape(pts: npt.NDArray[np.floating[Any]]) -> float:
    pts2 = np.roll(pts, 1, axis=0)
    dx = pts2[:, 0] - pts[:, 0]
    y = pts2[:, 1] + pts[:, 1]
    return float(np.sign((dx * y).sum()))


def area(pts: npt.NDArray[np.floating[Any]]) -> float:
    """Returns the area."""
    pts2 = np.roll(pts, 1, axis=0)
    dx = pts2[:, 0] - pts[:, 0]
    y = pts2[:, 1] + pts[:, 1]
    return float(np.sum(dx * y) / 2)


def centered_diff(a: npt.NDArray[np.floating[Any]]) -> npt.NDArray[np.floating[Any]]:
    d = (np.roll(a, -1, axis=0) - np.roll(a, 1, axis=0)) / 2
    return np.array(d)[1:-1]


def centered_diff2(a: npt.NDArray[np.floating[Any]]) -> npt.NDArray[np.floating[Any]]:
    d = (np.roll(a, -1, axis=0) - a) - (a - np.roll(a, 1, axis=0))
    return np.array(d[1:-1])


def curvature(
    points: npt.NDArray[np.floating[Any]], t: npt.NDArray[np.floating[Any]]
) -> npt.NDArray[np.floating[Any]]:
    """Args are the points and the tangents at each point.

        points : numpy.array shape (n, 2)
        t: numpy.array of size n

    Return:
        The curvature at each point.

    Computes the curvature at every point excluding the first and last point.

    For a planar curve parametrized as P(t) = (x(t), y(t)), the curvature is given
    by (x' y'' - x'' y' ) / (x' **2 + y' **2)**(3/2)

    """
    # Use centered difference for derivative
    dt = centered_diff(t)
    dp = centered_diff(points)
    dp2 = centered_diff2(points)

    dx = dp[:, 0] / dt
    dy = dp[:, 1] / dt

    dx2 = dp2[:, 0] / dt**2
    dy2 = dp2[:, 1] / dt**2

    res = (dx * dy2 - dx2 * dy) / (dx**2 + dy**2) ** (3 / 2)
    assert isinstance(res, np.ndarray)
    return res


def radius_of_curvature(
    points: npt.NDArray[np.floating[Any]], t: npt.NDArray[np.floating[Any]]
) -> npt.NDArray[np.floating[Any]]:
    return 1 / curvature(points, t)


def path_length(points: npt.NDArray[np.floating[Any]]) -> float:
    """Returns: The path length.

    Args:
        points: With shape (N, 2) representing N points with coordinates x, y.
    """
    dpts = points[1:, :] - points[:-1, :]
    _d = dpts**2
    return float(np.sum(np.sqrt(_d[:, 0] + _d[:, 1])))


def snap_angle(a: float) -> float:
    """Returns angle snapped along manhattan angle (0, 90, 180, 270).

    a: angle in deg
    Return angle snapped along manhattan angle
    """
    a = a % 360
    if -45 < a < 45:
        return 0
    elif 45 < a < 135:
        return 90
    elif 135 < a < 225:
        return 180
    elif 225 < a < 315:
        return 270
    else:
        return 0


def angles_rad(pts: npt.NDArray[np.floating[Any]]) -> npt.NDArray[np.floating[Any]]:
    """Returns the angles (radians) of the connection between each point and the next."""
    _pts = np.roll(pts, -1, 0)
    return np.array(np.arctan2(_pts[:, 1] - pts[:, 1], _pts[:, 0] - pts[:, 0]))


def angles_deg(pts: npt.NDArray[np.floating[Any]]) -> npt.NDArray[np.floating[Any]]:
    """Returns the angles (degrees) of the connection between each point and the next."""
    return angles_rad(pts) * RAD2DEG


def extrude_path(
    points: npt.NDArray[np.floating[Any]],
    width: float,
    with_manhattan_facing_angles: bool = True,
    spike_length: float64 | int | float = 0,
    start_angle: int | None = None,
    end_angle: int | None = None,
    grid: float | None = None,
) -> npt.NDArray[np.floating[Any]]:
    """Extrude a path of `width` along a curve defined by `points`.

    Args:
        points: numpy 2D array of shape (N, 2).
        width: of the path to extrude.
        with_manhattan_facing_angles: snaps to manhattan angles.
        spike_length: in um.
        start_angle: in degrees.
        end_angle: in degrees.
        grid: in um.

    Returns:
        numpy 2D array of shape (2*N, 2).
    """
    grid = grid or gf.kcl.dbu

    assert grid is not None

    if isinstance(points, list):
        points = np.stack([(p[0], p[1]) for p in points], axis=0)

    a = angles_deg(points)
    if with_manhattan_facing_angles:
        _start_angle = snap_angle(a[0] + 180)
        _end_angle = snap_angle(a[-2])
    else:
        _start_angle = a[0] + 180
        _end_angle = a[-2]

    start_angle_ = start_angle if start_angle is not None else _start_angle
    end_angle_ = end_angle if end_angle is not None else _end_angle

    assert start_angle_ is not None
    assert end_angle_ is not None

    a2 = angles_rad(points) * 0.5
    a1 = np.roll(a2, 1)

    a2[-1] = end_angle_ * DEG2RAD - a2[-2]
    a1[0] = start_angle_ * DEG2RAD - a1[1]

    a_plus = a2 + a1
    cos_a_min = np.cos(a2 - a1)
    offsets = np.column_stack((-sin(a_plus) / cos_a_min, cos(a_plus) / cos_a_min)) * (
        0.5 * width
    )

    points_back = np.flipud(points - offsets)
    if spike_length != 0:
        d = spike_length
        a_start = start_angle_ * DEG2RAD
        a_end = end_angle_ * DEG2RAD
        p_start_spike = points[0] + d * np.array([[cos(a_start), sin(a_start)]])
        p_end_spike = points[-1] + d * np.array([[cos(a_end), sin(a_end)]])

        pts = np.vstack((p_start_spike, points + offsets, p_end_spike, points_back))
    else:
        pts = np.vstack((points + offsets, points_back))

    return np.array(np.round(pts / grid) * grid)


def trim(
    component: Component,
    domain: Sequence[tuple[float, float]],
    flatten: bool = False,
) -> gf.Component:
    """Trim a component by another geometry, preserving the component's layers and ports.

    Useful to get a smaller component from a larger one for simulation.

    Args:
        component: Component(/Reference).
        domain: list of array-like[N][2] representing the boundary of the component to keep.
        flatten: if True, flattens the component.

    Returns: New component with layers (and possibly ports) of the component restricted to the domain.

    .. plot::
      :include-source:

      import gdsfactory as gf
      c = gf.components.straight_pin(length=10)
      trimmed_c = gf.functions.trim(component=c, domain=[[0, -5], [0, 5], [5, 5], [5, -5]])
      trimmed_c.plot()
    """
    dummy = gf.Component()
    dummy.add_polygon(domain, layer=(1, 0))
    dbbox = dummy.dbbox()
    left, bottom, right, top = dbbox.left, dbbox.bottom, dbbox.right, dbbox.top
    component.trim(left=left, right=right, bottom=bottom, top=top, flatten=flatten)
    return component


@gf.cell
def rotate(component: Component, angle: float) -> gf.Component:
    """Rotate a component by an angle in degrees.

    Args:
        component: to rotate.
        angle: in increments of 90°.

    Returns: Rotated component.
    """
    c = gf.Component()
    component = gf.get_component(component)
    ref = c.add_ref(component)
    ref.rotate(angle=angle)
    c.add_ports(ref.ports)
    c.copy_child_info(component)
    return c


rotate90 = partial(rotate, angle=90)
rotate180 = partial(rotate, angle=180)
rotate270 = partial(rotate, angle=270)


@gf.cell
def mirror(component: Component, x_mirror: bool = True) -> gf.Component:
    """Rotate a component by an angle in degrees.

    Args:
        component: to rotate.
        x_mirror: if True, mirrors the component along the x-axis.

    Returns: Rotated component.
    """
    c = gf.Component()
    component = gf.get_component(component)
    ref = c.add_ref(component)
    if x_mirror:
        ref.mirror_x()
    else:
        ref.mirror_y()
    c.add_ports(ref.ports)
    c.copy_child_info(component)
    return c



"""A minimal implementation of Dubins paths for waveguide routing adapted for gdsFactory by Quentin Wach.

https://quentinwach.com/blog/2024/02/15/dubins-paths-for-waveguide-routing.html
"""

import math as m
from math import cos, radians, sin
from typing import Literal

import kfactory as kf
from kfactory.routing.aa.optical import OpticalAllAngleRoute

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bends import bend_circular_all_angle
from gdsfactory.components.waveguides import straight_all_angle
from gdsfactory.typings import CrossSectionSpec, Port


def route_dubins(
    component: Component,
    port1: Port,
    port2: Port,
    cross_section: CrossSectionSpec,
) -> OpticalAllAngleRoute:
    """Route between ports using Dubins paths with radius from cross-section.

    Args:
        component: component to add the route to.
        port1: input port.
        port2: output port.
        cross_section: cross-section.
    """
    # Get start position and orientation
    x1, y1 = port1.center
    angle1 = float(port1.orientation)
    START = (x1, y1, angle1)  # Convert to um

    # Get end position and orientation
    x2, y2 = port2.center
    angle2 = float(port2.orientation)
    angle2 = (angle2 + 180) % 360  # Adjust for input connection
    END = (x2, y2, angle2)  # Convert to um

    xs = gf.get_cross_section(cross_section)
    # Find the Dubins path between ports using radius from cross-section
    path = dubins_path(start=START, end=END, cross_section=xs)  # Convert radius to um
    instances = place_dubins_path(component, xs, port1, solution=path)
    length = dubins_path_length(START, END, xs)

    backbone = [gf.kdb.DPoint(x1, y1), gf.kdb.DPoint(x2, y2)]  # TODO: fix this
    return OpticalAllAngleRoute(
        backbone=backbone,
        start_port=port1.to_itype(),
        end_port=port2.to_itype(),
        length=length,
        instances=instances,
    )


def general_planner(
    planner: str, alpha: float, beta: float, d: float
) -> tuple[list[float | Literal[0]], list[str], float] | None:
    """Finds the optimal path between two points using various planning methods."""
    sa = m.sin(alpha)
    sb = m.sin(beta)
    ca = m.cos(alpha)
    cb = m.cos(beta)
    c_ab = m.cos(alpha - beta)
    mode = list(planner)

    planner_uc = planner.upper()

    if planner_uc == "LSL":
        tmp0 = d + sa - sb
        p_squared = 2 + (d * d) - (2 * c_ab) + (2 * d * (sa - sb))
        if p_squared < 0:
            return None
        tmp1 = m.atan2((cb - ca), tmp0)
        t = mod_to_pi(-alpha + tmp1)
        p = m.sqrt(p_squared)
        q = mod_to_pi(beta - tmp1)

    elif planner_uc == "RSR":
        tmp0 = d - sa + sb
        p_squared = 2 + (d * d) - (2 * c_ab) + (2 * d * (sb - sa))
        if p_squared < 0:
            return None
        tmp1 = m.atan2((ca - cb), tmp0)
        t = mod_to_pi(alpha - tmp1)
        p = m.sqrt(p_squared)
        q = mod_to_pi(-beta + tmp1)

    elif planner_uc == "LSR":
        p_squared = -2 + (d * d) + (2 * c_ab) + (2 * d * (sa + sb))
        if p_squared < 0:
            return None
        p = m.sqrt(p_squared)
        tmp2 = m.atan2((-ca - cb), (d + sa + sb)) - m.atan2(-2.0, p)
        t = mod_to_pi(-alpha + tmp2)
        q = mod_to_pi(-mod_to_pi(beta) + tmp2)

    elif planner_uc == "RSL":
        p_squared = (d * d) - 2 + (2 * c_ab) - (2 * d * (sa + sb))
        if p_squared < 0:
            return None
        p = m.sqrt(p_squared)
        tmp2 = m.atan2((ca + cb), (d - sa - sb)) - m.atan2(2.0, p)
        t = mod_to_pi(alpha - tmp2)
        q = mod_to_pi(beta - tmp2)

    elif planner_uc == "RLR":
        tmp_rlr = (6.0 - d * d + 2.0 * c_ab + 2.0 * d * (sa - sb)) / 8.0
        if abs(tmp_rlr) > 1.0:
            return None

        p = mod_to_pi(2 * m.pi - m.acos(tmp_rlr))
        t = mod_to_pi(alpha - m.atan2(ca - cb, d - sa + sb) + mod_to_pi(p / 2.0))
        q = mod_to_pi(alpha - beta - t + mod_to_pi(p))

    elif planner_uc == "LRL":
        tmp_lrl = (6.0 - d * d + 2 * c_ab + 2 * d * (-sa + sb)) / 8.0
        if abs(tmp_lrl) > 1:
            return None
        p = mod_to_pi(2 * m.pi - m.acos(tmp_lrl))
        t = mod_to_pi(-alpha - m.atan2(ca - cb, d + sa - sb) + p / 2.0)
        q = mod_to_pi(mod_to_pi(beta) - alpha - t + mod_to_pi(p))

    else:
        raise ValueError(f"Invalid planner: {planner}")

    path = [t, p, q]

    for i in [0, 2]:
        if planner[i].islower():
            path[i] = (2 * m.pi) - path[i]

    cost = sum(map(abs, path))

    return (path, mode, cost)


def dubins_path_length(
    start: tuple[float, float, float],
    end: tuple[float, float, float],
    xs: CrossSectionSpec,
) -> float:
    """Calculate the length of a Dubins path."""
    (sx, sy, syaw) = start
    (ex, ey, eyaw) = end
    # convert the degree angle inputs to radians
    syaw = m.radians(syaw)
    eyaw = m.radians(eyaw)

    ex = ex - sx
    ey = ey - sy

    lex = m.cos(syaw) * ex + m.sin(syaw) * ey
    ley = -m.sin(syaw) * ex + m.cos(syaw) * ey
    return m.sqrt(lex**2.0 + ley**2.0)


def dubins_path(
    start: tuple[float, float, float],
    end: tuple[float, float, float],
    cross_section: CrossSectionSpec,
) -> list[tuple[str, float, float]]:
    """Finds the Dubins path between two points."""
    xs = gf.get_cross_section(cross_section)
    (sx, sy, syaw) = start  # Coordinates already in um
    (ex, ey, eyaw) = end  # Coordinates already in um

    # Convert angles to radians
    syaw = m.radians(syaw)
    eyaw = m.radians(eyaw)

    # Use radius in um
    c = xs.radius  # Already converted to um

    assert c is not None, "Cross-section radius is None"

    # Calculate relative end position
    ex = ex - sx
    ey = ey - sy

    # Transform to local coordinates
    lex = m.cos(syaw) * ex + m.sin(syaw) * ey
    ley = -m.sin(syaw) * ex + m.cos(syaw) * ey
    leyaw = eyaw - syaw

    # Calculate normalized distance
    D = m.sqrt(lex**2.0 + ley**2.0)
    d = D / c  # Normalize by radius

    # Calculate angles for path planning
    theta = mod_to_pi(m.atan2(ley, lex))
    alpha = mod_to_pi(-theta)
    beta = mod_to_pi(leyaw - theta)

    # Find best path
    planners = ["LSL", "RSR", "LSR", "RSL", "RLR", "LRL"]
    bcost = float("inf")
    bt, bp, bq, bmode = None, None, None, None

    for planner in planners:
        solution = general_planner(planner, alpha, beta, d)
        if solution is None:
            continue
        (path, mode, cost) = solution
        (t, p, q) = path
        if bcost > cost:
            bt, bp, bq, bmode = t, p, q, mode
            bcost = cost

    assert bt is not None and bp is not None and bq is not None and bmode is not None

    # Return path segments with lengths in um
    return list(zip(bmode, [bt * c, bp * c, bq * c], [c] * 3, strict=False))


def mod_to_pi(angle: float) -> float:
    """Normalizes an angle to the range [0, 2*pi)."""
    return angle - 2.0 * m.pi * m.floor(angle / 2.0 / m.pi)


def pi_to_pi(angle: float) -> float:
    """Constrains an angle to the range [-pi, pi]."""
    while angle >= m.pi:
        angle = angle - 2.0 * m.pi
    while angle <= -m.pi:
        angle = angle + 2.0 * m.pi
    return angle


def linear(
    start: tuple[float, float, float], end: tuple[float, float, float], steps: int
) -> tuple[list[float], list[float]]:
    """Creates a list of points on lines between a given start point and end point.

    start/end: [x, y, angle], the start/end point with given jaw angle.
    """
    start_x, start_y = start[0], start[1]
    dx = (end[0] - start_x) / steps
    dy = (end[1] - start_y) / steps

    # Preallocate output array
    x = [start_x + i * dx for i in range(steps + 1)]
    y = [start_y + i * dy for i in range(steps + 1)]
    return x, y


def arrow_orientation(angle: float) -> tuple[float, float]:
    """Returns x, y setoffs for a given angle to orient the arrows marking the yaw of the start and end points."""
    rad = radians(angle)
    alpha_x = cos(rad)
    alpha_y = sin(rad)
    return alpha_x, alpha_y


def place_dubins_path(
    component: Component,
    xs: CrossSectionSpec,
    port1: Port,
    solution: list[tuple[str, float, float]],
) -> list[kf.VInstance]:
    """Creates GDS component with Dubins path.

    Args:
        component: component to add the route to.
        xs: cross-section.
        port1: input port.
        solution: Dubins path solution.
    """
    c = component
    current_position = port1

    instances: list[kf.VInstance] = []

    for mode, length, radius in solution:
        if mode == "L":
            # Length and radius are in um, convert to nm for gdsfactory
            arc_angle = 180 * length / (m.pi * radius)
            bend = c.add_ref_off_grid(
                bend_circular_all_angle(angle=arc_angle, cross_section=xs)  # type: ignore[arg-type]
            )
            bend.connect("o1", current_position)
            current_position = bend.ports["o2"]
            instances.append(bend)

        elif mode == "R":
            arc_angle = -(180 * length / (m.pi * radius))
            bend = c.add_ref_off_grid(
                bend_circular_all_angle(angle=arc_angle, cross_section=xs)  # type: ignore[arg-type]
            )
            bend.connect("o1", current_position)
            current_position = bend.ports["o2"]
            instances.append(bend)

        elif mode == "S":
            straight = c.add_ref_off_grid(
                straight_all_angle(length=length, cross_section=xs)  # type: ignore[arg-type]
            )
            straight.connect("o1", current_position)
            current_position = straight.ports["o2"]
            instances.append(straight)

        else:
            raise ValueError(f"Invalid mode: {mode}")

    return instances



from collections.abc import Sequence
from typing import Any

import numpy as np
import numpy.typing as npt

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.cross_section import Section
from gdsfactory.path import Path
from gdsfactory.typings import LayerSpec


@gf.cell_with_module_name
def straight_piecewise(
    x: Sequence[float] | Path,
    widths: Sequence[float],
    layer: LayerSpec,
    sections: Sequence[Section] | None = None,
    port_names: tuple[str | None, str | None] = ("o1", "o2"),
    name: str = "core",
    **kwargs: Any,
) -> Component:
    """Create a component with a piecewise-defined straight waveguide.

    Args:
        x: X coordinates or a custom Path object.
        widths: Waveguide widths at each corresponding x.
        layer: Layer to extrude.
        sections: Additional cross-section sections to extrude.
        port_names: Port names for the waveguide.
        name: Name for the core (main) Section.
        **kwargs: Additional keyword arguments for the Section.
    """
    if isinstance(x, Sequence) and len(x) != len(widths):
        raise ValueError("x and widths must have the same length.")

    def width_function(_: float) -> npt.NDArray[np.float64]:
        return np.array(widths)

    if isinstance(x, gf.Path):
        p = x
    else:
        p = gf.Path()
        p.points = np.array([(xi, 0.0) for xi in x])

    section_list = list(sections or [])
    section_list.append(
        Section(
            name=name,
            width=0,
            width_function=width_function,
            offset=0,
            layer=layer,
            port_names=port_names,
            **kwargs,
        )
    )
    cross_section = gf.CrossSection(sections=tuple(section_list))

    return gf.path.extrude(p, cross_section=cross_section)



"""GDS regression test. Inspired by lytest."""

import filecmp
import pathlib
import shutil

import kfactory as kf
from kfactory import DKCell, KCLayout, kdb, logger

import gdsfactory as gf
from gdsfactory.config import CONF, PATH
from gdsfactory.name import clean_name, get_name_short


class GeometryDifference(Exception):
    pass


PathType = pathlib.Path | str


def xor(
    old: KCLayout,
    new: KCLayout,
    test_name: str = "",
    ignore_sliver_differences: bool | None = None,
    ignore_cell_name_differences: bool | None = None,
    ignore_label_differences: bool | None = None,
    stagger: bool = True,
) -> DKCell:
    """Returns XOR of two layouts.

    Args:
        old: reference layout.
        new: run layout.
        test_name: prefix for the new cell.
        ignore_sliver_differences: if True, ignores any sliver differences in the XOR result. If None (default), defers to the value set in CONF.difftest_ignore_sliver_differences
        ignore_cell_name_differences: if True, ignores any cell name differences. If None (default), defers to the value set in CONF.difftest_ignore_cell_name_differences
        ignore_label_differences: if True, ignores any label differences when run in XOR mode. If None (default) defers to the value set in CONF.difftest_ignore_label_differences
        stagger: if True, staggers the old/new/xor views. If False, all three are overlaid.
    """
    if ignore_sliver_differences is None:
        ignore_sliver_differences = CONF.difftest_ignore_sliver_differences

    if ignore_cell_name_differences is None:
        ignore_cell_name_differences = CONF.difftest_ignore_cell_name_differences

    if ignore_label_differences is None:
        ignore_label_differences = CONF.difftest_ignore_label_differences

    if old.kcl.dbu != new.kcl.dbu:
        raise ValueError(
            f"dbu is different in old {old.kcl.dbu} and new {new.kcl.dbu} cells"
        )

    equivalent = True
    ld = kdb.LayoutDiff()

    a_regions: dict[int, kdb.Region] = {}
    a_texts: dict[int, kdb.Texts] = {}
    b_regions: dict[int, kdb.Region] = {}
    b_texts: dict[int, kdb.Texts] = {}

    def get_region(key: int, regions: dict[int, kdb.Region]) -> kdb.Region:
        if key not in regions:
            reg = kdb.Region()
            regions[key] = reg
            return reg
        else:
            return regions[key]

    def get_texts(key: int, texts_dict: dict[int, kdb.Texts]) -> kdb.Texts:
        if key not in texts_dict:
            texts = kdb.Texts()
            texts_dict[key] = texts
            return texts
        else:
            return texts_dict[key]

    def polygon_diff_a(anotb: kdb.Polygon, prop_id: int) -> None:
        get_region(ld.layer_index_a(), a_regions).insert(anotb)

    def polygon_diff_b(bnota: kdb.Polygon, prop_id: int) -> None:
        get_region(ld.layer_index_b(), b_regions).insert(bnota)

    def cell_diff_a(cell: kdb.Cell) -> None:
        nonlocal equivalent
        print(f"{cell.name} only in old")
        if not ignore_cell_name_differences:
            equivalent = False

    def cell_diff_b(cell: kdb.Cell) -> None:
        nonlocal equivalent
        print(f"{cell.name} only in new")
        if not ignore_cell_name_differences:
            equivalent = False

    def text_diff_a(anotb: kdb.Text, prop_id: int) -> None:
        print("Text only in old")
        get_texts(ld.layer_index_a(), a_texts).insert(anotb)

    def text_diff_b(bnota: kdb.Text, prop_id: int) -> None:
        print("Text only in new")
        get_texts(ld.layer_index_b(), b_texts).insert(bnota)

    ld.on_cell_in_a_only = lambda anotb: cell_diff_a(anotb)  # type: ignore[assignment]
    ld.on_cell_in_b_only = lambda anotb: cell_diff_b(anotb)  # type: ignore[assignment]
    ld.on_text_in_a_only = lambda anotb, prop_id: text_diff_a(anotb, prop_id)  # type: ignore[assignment]
    ld.on_text_in_b_only = lambda anotb, prop_id: text_diff_b(anotb, prop_id)  # type: ignore[assignment]

    ld.on_polygon_in_a_only = lambda anotb, prop_id: polygon_diff_a(anotb, prop_id)  # type: ignore[assignment]
    ld.on_polygon_in_b_only = lambda anotb, prop_id: polygon_diff_b(anotb, prop_id)  # type: ignore[assignment]

    if ignore_cell_name_differences:
        ld.on_cell_name_differs = lambda anotb: print(f"cell name differs {anotb.name}")  # type: ignore[assignment]
        equal = ld.compare(
            old.kdb_cell,
            new.kdb_cell,
            kdb.LayoutDiff.SmartCellMapping | kdb.LayoutDiff.Verbose,
            1,
        )
    else:
        equal = ld.compare(old.kdb_cell, new.kdb_cell, kdb.LayoutDiff.Verbose, 1)

    if not ignore_label_differences and (a_texts or b_texts):
        equivalent = False
    if equal:
        return gf.Component(name="xor_empty")
    c = DKCell(name=f"{test_name}_difftest")
    ref = old
    run = new

    old_kcell = DKCell(name=f"{test_name}_old")
    new_kcell = DKCell(name=f"{test_name}_new")

    old_kcell.copy_tree(ref.kdb_cell)
    new_kcell.copy_tree(run.kdb_cell)

    old_kcell.name = f"{test_name}_old"
    new_kcell.name = f"{test_name}_new"

    old_ref = c << old_kcell
    new_ref = c << new_kcell

    dy = 10
    if stagger:
        old_ref.movey(+old_kcell.ysize + dy)
        new_ref.movey(-old_kcell.ysize - dy)

    layer_label = kf.kcl.layout.layer(1, 0)
    c.shapes(layer_label).insert(kf.kdb.DText("old", old_ref.dtrans))
    c.shapes(layer_label).insert(kf.kdb.DText("new", new_ref.dtrans))
    c.shapes(layer_label).insert(
        kf.kdb.DText(
            "xor", kf.kdb.DTrans(new_ref.xmin, old_ref.ymax - old_ref.ysize - dy)
        )
    )

    print("Running XOR on differences...")
    # assume equivalence until we find XOR differences, determined significant by the settings
    diff = DKCell(name=f"{test_name}_xor")

    for layer in c.kcl.layer_infos():
        # exists in both
        if (
            new_kcell.kcl.layout.find_layer(layer) is not None
            and old_kcell.kcl.layout.find_layer(layer) is not None
        ):
            layer_ref = old_kcell.layer(layer)
            layer_run = new_kcell.layer(layer)

            region_run = kdb.Region(new_kcell.begin_shapes_rec(layer_run))
            region_ref = kdb.Region(old_kcell.begin_shapes_rec(layer_ref))
            region_diff = region_run ^ region_ref

            if not region_diff.is_empty():
                layer_id = c.layer(layer)
                region_xor = region_ref ^ region_run
                diff.shapes(layer_id).insert(region_xor)
                xor_w_tolerance = region_xor.sized(-1)
                is_sliver = xor_w_tolerance.is_empty()
                message = f"{test_name}: XOR difference on layer {layer}"
                if is_sliver:
                    message += " (sliver)"
                    if not ignore_sliver_differences:
                        equivalent = False
                else:
                    equivalent = False
                print(message)
        # only in new
        elif new_kcell.kcl.layout.find_layer(layer) is not None:
            layer_id = new_kcell.layer(layer)
            region = kdb.Region(new_kcell.begin_shapes_rec(layer_id))
            diff.shapes(c.kcl.layer(layer)).insert(region)
            print(f"{test_name}: layer {layer} only exists in updated cell")
            equivalent = False

        # only in old
        elif old_kcell.kcl.layout.find_layer(layer) is not None:
            layer_id = old_kcell.layer(layer)
            region = kdb.Region(old_kcell.begin_shapes_rec(layer_id))
            diff.shapes(c.kcl.layer(layer)).insert(region)
            print(f"{test_name}: layer {layer} missing from updated cell")
            equivalent = False

        _ = c << diff
    return c


def diff(
    ref_file: PathType,
    run_file: PathType,
    xor: bool = True,
    test_name: str = "",
    ignore_sliver_differences: bool | None = None,
    ignore_cell_name_differences: bool | None = None,
    ignore_label_differences: bool | None = None,
    show: bool = False,
    stagger: bool = True,
    out_file: PathType | None = None,
) -> bool:
    """Returns True if files are different, prints differences and shows them in klayout.

    Args:
        ref_file: reference (old) file.
        run_file: run (new) file.
        xor: runs xor on every layer between old and run files.
        test_name: prefix for the new cell.
        ignore_sliver_differences: if True, ignores any sliver differences in the XOR result. If None (default), defers to the value set in CONF.difftest_ignore_sliver_differences
        ignore_cell_name_differences: if True, ignores any cell name differences. If None (default), defers to the value set in CONF.difftest_ignore_cell_name_differences
        ignore_label_differences: if True, ignores any label differences when run in XOR mode. If None (default) defers to the value set in CONF.difftest_ignore_label_differences
        show: shows diff in klayout.
        stagger: if True, staggers the old/new/xor views. If False, all three are overlaid.
        out_file: if not None, saves the diff to the specified file.
    """
    ref_file, run_file = pathlib.Path(ref_file), pathlib.Path(run_file)
    if ref_file == run_file:
        return False

    old = read_top_cell(ref_file)
    new = read_top_cell(run_file)

    if ignore_sliver_differences is None:
        ignore_sliver_differences = CONF.difftest_ignore_sliver_differences

    if ignore_cell_name_differences is None:
        ignore_cell_name_differences = CONF.difftest_ignore_cell_name_differences

    if ignore_label_differences is None:
        ignore_label_differences = CONF.difftest_ignore_label_differences

    if old.kcl.dbu != new.kcl.dbu:
        raise ValueError(
            f"dbu is different in old {old.kcl.dbu} {ref_file} and new {new.kcl.dbu} {run_file} files"
        )

    equivalent = True
    ld = kdb.LayoutDiff()

    a_regions: dict[int, kdb.Region] = {}
    a_texts: dict[int, kdb.Texts] = {}
    b_regions: dict[int, kdb.Region] = {}
    b_texts: dict[int, kdb.Texts] = {}

    def get_region(key: int, regions: dict[int, kdb.Region]) -> kdb.Region:
        if key not in regions:
            reg = kdb.Region()
            regions[key] = reg
            return reg
        else:
            return regions[key]

    def get_texts(key: int, texts_dict: dict[int, kdb.Texts]) -> kdb.Texts:
        if key not in texts_dict:
            texts = kdb.Texts()
            texts_dict[key] = texts
            return texts
        else:
            return texts_dict[key]

    def polygon_diff_a(anotb: kdb.Polygon, prop_id: int) -> None:
        get_region(ld.layer_index_a(), a_regions).insert(anotb)

    def polygon_diff_b(bnota: kdb.Polygon, prop_id: int) -> None:
        get_region(ld.layer_index_b(), b_regions).insert(bnota)

    def cell_diff_a(cell: kdb.Cell) -> None:
        nonlocal equivalent
        print(f"{cell.name} only in old")
        if not ignore_cell_name_differences:
            equivalent = False

    def cell_diff_b(cell: kdb.Cell) -> None:
        nonlocal equivalent
        print(f"{cell.name} only in new")
        if not ignore_cell_name_differences:
            equivalent = False

    def text_diff_a(anotb: kdb.Text, prop_id: int) -> None:
        print("Text only in old")
        get_texts(ld.layer_index_a(), a_texts).insert(anotb)

    def text_diff_b(bnota: kdb.Text, prop_id: int) -> None:
        print("Text only in new")
        get_texts(ld.layer_index_b(), b_texts).insert(bnota)

    ld.on_cell_in_a_only = cell_diff_a  # type: ignore[assignment]
    ld.on_cell_in_b_only = cell_diff_b  # type: ignore[assignment]
    ld.on_text_in_a_only = text_diff_a  # type: ignore[assignment]
    ld.on_text_in_b_only = text_diff_b  # type: ignore[assignment]

    ld.on_polygon_in_a_only = polygon_diff_a  # type: ignore[assignment]
    ld.on_polygon_in_b_only = polygon_diff_b  # type: ignore[assignment]

    if ignore_cell_name_differences:
        ld.on_cell_name_differs = lambda anotb: print(f"cell name differs {anotb.name}")  # type: ignore[assignment]
        equal = ld.compare(
            old.kdb_cell,
            new.kdb_cell,
            kdb.LayoutDiff.SmartCellMapping | kdb.LayoutDiff.Verbose,
            1,
        )
    else:
        equal = ld.compare(old.kdb_cell, new.kdb_cell, kdb.LayoutDiff.Verbose, 1)

    if not ignore_label_differences and (a_texts or b_texts):
        equivalent = False

    if not equal:
        c = DKCell(name=f"{test_name}_difftest")
        ref = old
        run = new

        old = DKCell(name=f"{test_name}_old")
        new = DKCell(name=f"{test_name}_new")

        old.copy_tree(ref.kdb_cell)
        new.copy_tree(run.kdb_cell)

        old.name = f"{test_name}_old"
        new.name = f"{test_name}_new"

        old_ref = c << old
        new_ref = c << new

        dy = 10
        if stagger:
            old_ref.movey(+old.ysize + dy)
            new_ref.movey(-old.ysize - dy)

        layer_label = kf.kcl.layout.layer(1, 0)
        c.shapes(layer_label).insert(kf.kdb.DText("old", old_ref.dtrans))
        c.shapes(layer_label).insert(kf.kdb.DText("new", new_ref.dtrans))
        if xor:
            c.shapes(layer_label).insert(
                kf.kdb.DText(
                    "xor",
                    kf.kdb.DTrans(new_ref.xmin, old_ref.ymax - old_ref.ysize - dy),
                )
            )

        if xor:
            print("Running XOR on differences...")
            # assume equivalence until we find XOR differences, determined significant by the settings
            diff = DKCell(name=f"{test_name}_xor")

            for layer in c.kcl.layer_infos():
                # exists in both
                if (
                    new.kcl.layout.find_layer(layer) is not None
                    and old.kcl.layout.find_layer(layer) is not None
                ):
                    layer_ref = old.layer(layer)
                    layer_run = new.layer(layer)

                    region_run = kdb.Region(new.begin_shapes_rec(layer_run))
                    region_ref = kdb.Region(old.begin_shapes_rec(layer_ref))
                    region_diff = region_run ^ region_ref

                    if not region_diff.is_empty():
                        layer_id = c.layer(layer)
                        region_xor = region_ref ^ region_run
                        diff.shapes(layer_id).insert(region_xor)
                        xor_w_tolerance = region_xor.sized(-1)
                        is_sliver = xor_w_tolerance.is_empty()
                        message = f"{test_name}: XOR difference on layer {layer}"
                        if is_sliver:
                            message += " (sliver)"
                            if not ignore_sliver_differences:
                                equivalent = False
                        else:
                            equivalent = False
                        print(message)
                # only in new
                elif new.kcl.layout.find_layer(layer) is not None:
                    layer_id = new.layer(layer)
                    region = kdb.Region(new.begin_shapes_rec(layer_id))
                    diff.shapes(c.kcl.layer(layer)).insert(region)
                    print(f"{test_name}: layer {layer} only exists in updated cell")
                    equivalent = False

                # only in old
                elif old.kcl.layout.find_layer(layer) is not None:
                    layer_id = old.layer(layer)
                    region = kdb.Region(old.begin_shapes_rec(layer_id))
                    diff.shapes(c.kcl.layer(layer)).insert(region)
                    print(f"{test_name}: layer {layer} missing from updated cell")
                    equivalent = False

            _ = c << diff
            if equivalent:
                print("No significant XOR differences between layouts!")
        else:
            # if no additional xor verification, the two files are not equivalent
            equivalent = False

        if not equivalent:
            if out_file is not None:
                c.write(out_file)
            if show:
                c.show()

        return not equivalent
    return False


def difftest(
    component: gf.Component,
    test_name: str | None = None,
    dirpath: pathlib.Path = PATH.gds_ref,
    xor: bool = True,
    dirpath_run: pathlib.Path = PATH.gds_run,
    ignore_sliver_differences: bool | None = None,
) -> None:
    """Avoids GDS regressions tests on the GeometryDifference.

    If files are the same it returns None. If files are different runs XOR
    between new component and the GDS reference stored in dirpath and
    raises GeometryDifference if there are differences and show differences in KLayout.

    If it runs for the fist time it just stores the GDS reference.

    Args:
        component: to test if it has changed.
        test_name: used to store the GDS file.
        dirpath: directory where reference files are stored.
        xor: runs XOR.
        dirpath_run: directory to store gds file generated by the test.
        ignore_sliver_differences: if True, ignores any sliver differences in the XOR result.
            If None (default), defers to the value set in CONF.difftest_ignore_sliver_differences
    """
    test_name = test_name or (
        f"{component.function_name}_{component.name}"
        if hasattr(component, "function_name")
        and component.name != component.function_name
        else f"{component.name}"
    )
    filename = f"{test_name}.gds"
    dirpath_ref = dirpath
    dirpath_ref.mkdir(exist_ok=True, parents=True)
    dirpath_run.mkdir(exist_ok=True, parents=True)

    filename = get_name_short(clean_name(test_name))

    ref_file = dirpath_ref / f"{filename}.gds"
    run_file = dirpath_run / f"{filename}.gds"

    component = gf.get_component(component)
    run_file = component.write_gds(gdspath=run_file)

    if not ref_file.exists():
        shutil.copy(run_file, ref_file)
        raise AssertionError(
            f"Reference GDS file for {test_name!r} not found. Writing to {ref_file!r}"
        )

    if filecmp.cmp(ref_file, run_file, shallow=False):
        return

    if diff(
        ref_file=ref_file,
        run_file=run_file,
        xor=xor,
        test_name=test_name,
        ignore_sliver_differences=ignore_sliver_differences,
        show=True,
    ):
        print(
            f"\ngds_run {filename!r} changed from gds_ref {str(ref_file)!r}\n"
            "You can check the differences in Klayout GUI or run XOR with\n"
            f"gf gds-diff --xor {ref_file} {run_file}\n"
        )
        try:
            overwrite(ref_file, run_file)
        except OSError as exc:
            raise GeometryDifference(
                "\n"
                f"{filename!r} changed from reference {str(ref_file)!r}. "
                "Run `pytest -s` to step and check differences in klayout GUI."
            ) from exc


def overwrite(ref_file: pathlib.Path, run_file: pathlib.Path) -> None:
    val = input("Save current GDS as the new reference (Y)? [Y/n]")
    if val.upper().startswith("N"):
        raise GeometryDifference

    logger.info(f"deleting file {str(ref_file)!r}")
    ref_file.unlink()
    shutil.copy(run_file, ref_file)
    raise GeometryDifference


def read_top_cell(arg0: pathlib.Path) -> kf.DKCell:
    kcl = KCLayout(name=str(arg0))
    kcl.read(arg0)
    kcell = kcl.dkcells[kcl.top_cell().name]

    if hasattr(kcl, "cross_sections"):
        for cross_section in kcl.cross_sections.cross_sections.values():
            kf.kcl.get_symmetrical_cross_section(cross_section)
    return kcell



from __future__ import annotations

from typing import Any

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.typings import ComponentSpec, LayerSpecs


@gf.cell_with_module_name
def copy_layers(
    factory: ComponentSpec = "cross",
    layers: LayerSpecs = ((1, 0), (2, 0)),
    flatten: bool = False,
    **kwargs: Any,
) -> Component:
    """Returns a component with the geometry copied in different layers.

    Args:
        factory: component spec.
        layers: iterable of layers.
        flatten: flatten the result.
        kwargs: keyword arguments passed to the component.
    """
    c = Component()

    ci = None
    for layer in layers:
        c << (ci := gf.get_component(factory, layer=layer, **kwargs))
    if ci is not None:
        c.copy_child_info(ci)

    if flatten:
        c.flatten()
    return c



from __future__ import annotations

from functools import partial

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.tapers.taper import taper
from gdsfactory.typings import ComponentSpec

_terminator_function = partial(taper, width2=0.1)


@gf.cell_with_module_name
def add_termination(
    component: ComponentSpec = "straight",
    port_names: tuple[str, ...] | None = None,
    terminator: ComponentSpec = _terminator_function,
    terminator_port_name: str | None = None,
) -> Component:
    """Returns component with terminator on some ports.

    Args:
        component: to add terminator.
        port_names: ports to add terminator.
        terminator: factory for the terminator.
        terminator_port_name: for the terminator to connect to the component ports.
    """
    terminator = gf.get_component(terminator)
    terminator_port_name = terminator_port_name or terminator.ports[0].name

    assert terminator_port_name is not None

    c = Component()
    component = gf.get_component(component)
    ref = c.add_ref(component)

    ports_names_all = [p.name for p in component.ports]
    ports_names_to_terminate = port_names or ports_names_all

    for port_name in ports_names_all:
        if port_name in ports_names_to_terminate:
            t_ref = c.add_ref(terminator)
            t_ref.connect(port=terminator_port_name, other=ref[port_name])
        else:
            port = ref[port_name]
            c.add_port(name=port.name, port=port)

    c.copy_child_info(component)
    return c



from __future__ import annotations

from collections.abc import Sequence

from gdsfactory.typings import Port


class RouteWarning(UserWarning):
    pass


def direction_ports_from_list_ports(
    optical_ports: Sequence[Port],
) -> dict[str, list[Port]]:
    """Returns a dict of WENS ports."""
    direction_ports: dict[str, list[Port]] = {x: [] for x in ["E", "N", "W", "S"]}
    for p in optical_ports:
        orientation = (p.orientation + 360.0) % 360
        if orientation <= 45.0 or orientation >= 315:
            direction_ports["E"].append(p)
        elif orientation <= 135.0 and orientation >= 45.0:
            direction_ports["N"].append(p)
        elif orientation <= 225.0 and orientation >= 135.0:
            direction_ports["W"].append(p)
        else:
            direction_ports["S"].append(p)

    for direction, list_ports in list(direction_ports.items()):
        if direction in ["E", "W"]:
            list_ports.sort(key=lambda p: p.y)

        if direction in ["S", "N"]:
            list_ports.sort(key=lambda p: p.x)

    return direction_ports


def check_ports_have_equal_spacing(list_ports: Sequence[Port]) -> float:
    """Returns port separation.

    Raises error if not constant.

    """
    if not isinstance(list_ports, list):
        raise ValueError(f"list_ports should be a list of ports, got {list_ports}")
    if not list_ports:
        raise ValueError("list_ports should not be empty")

    orientation = get_list_ports_angle(list_ports)
    if orientation in [0, 180]:
        xys = [p.y for p in list_ports]
    else:
        xys = [p.x for p in list_ports]

    seps = [round(abs(c2 - c1), 5) for c1, c2 in zip(xys[1:], xys[:-1], strict=False)]
    different_seps = set(seps)
    if len(different_seps) > 1:
        raise ValueError(f"Ports should have the same separation. Got {different_seps}")
    return float(seps[0])


def get_list_ports_angle(list_ports: Sequence[Port]) -> float | None:
    """Returns the orientation/angle (in degrees) of a list of ports."""
    if not list_ports:
        return None
    if len({p.orientation for p in list_ports}) > 1:
        raise ValueError(f"All port angles should be the same. Got {list_ports}")
    return list_ports[0].orientation



"""# Flattening a Component.

Sometimes you want to remove cell structure from a Component while keeping all
of the shapes/polygons intact and in place.

The Component.flatten() method flattens current Component by copying all polygons from the underlying references.

"""

from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component


@gf.cell
def flatten_device() -> Component:
    c = gf.Component()

    c.add_ref(gf.components.rectangle(size=(10, 1), layer=(1, 0)))
    c.add_ref(gf.components.rectangle(size=(10, 2), layer=(3, 0)))
    c.add_ref(gf.components.rectangle(size=(10, 3), layer=(2, 0)))

    assert len(c.insts) == 3
    c.flatten()
    assert len(c.insts) == 0
    return c


if __name__ == "__main__":
    c = flatten_device()
    c.show()



from __future__ import annotations

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.superconductors.optimal_hairpin import optimal_hairpin
from gdsfactory.typings import LayerSpec, Port, Size


@gf.cell_with_module_name
def snspd(
    wire_width: float = 0.2,
    wire_pitch: float = 0.6,
    size: Size = (10, 8),
    num_squares: int | None = None,
    turn_ratio: float = 4,
    terminals_same_side: bool = False,
    layer: LayerSpec = (1, 0),
    port_type: str = "electrical",
) -> Component:
    """Creates an optimally-rounded SNSPD.

    Args:
        wire_width: Width of the wire.
        wire_pitch: Distance between two adjacent wires. Must be greater than `width`.
        size: Float2
            (width, height) of the rectangle formed by the outer boundary of the
            SNSPD.
        num_squares: int | None = None
            Total number of squares inside the SNSPD length.
        turn_ratio: float
            Specifies how much of the SNSPD width is dedicated to the 180 degree
            turn. A `turn_ratio` of 10 will result in 20% of the width being
            comprised of the turn.
        terminals_same_side: If True, both ports will be located on the same side of the SNSPD.
        layer: layer spec to put polygon geometry on.
        port_type: type of port to add to the component.

    """
    if num_squares is not None:
        xy = np.sqrt(num_squares * wire_pitch * wire_width)
        size = (xy, xy)
        num_squares = None

    xsize, ysize = size
    if num_squares is not None:
        if xsize is None:
            xsize = num_squares * wire_pitch * wire_width / ysize
        elif ysize is None:
            ysize = num_squares * wire_pitch * wire_width / xsize

    num_meanders = int(np.ceil(ysize / wire_pitch))

    D = Component()
    hairpin = optimal_hairpin(
        width=wire_width,
        pitch=wire_pitch,
        turn_ratio=turn_ratio,
        length=xsize / 2,
        num_pts=20,
        layer=layer,
    )

    if not terminals_same_side and (num_meanders % 2) == 0:
        num_meanders += 1
    elif terminals_same_side and (num_meanders % 2) == 1:
        num_meanders += 1

    port_type = "electrical"

    start_nw = D.add_ref(
        gf.c.compass(size=(xsize / 2, wire_width), layer=layer, port_type=port_type)
    )
    hp_prev = D.add_ref(hairpin)
    hp_prev.connect("e1", start_nw.ports["e3"])
    alternate = True
    last_port: Port | None = None
    for _n in range(2, num_meanders):
        hp = D.add_ref(hairpin)
        if alternate:
            hp.connect("e2", hp_prev.ports["e2"])
        else:
            hp.connect("e1", hp_prev.ports["e1"])
        last_port = hp.ports["e2"] if terminals_same_side else hp.ports["e1"]
        hp_prev = hp
        alternate = not alternate

    finish_se = D.add_ref(
        gf.c.compass(size=(xsize / 2, wire_width), layer=layer, port_type=port_type)
    )
    if last_port is not None:
        finish_se.connect("e3", last_port)

    D.add_port(port=start_nw.ports["e1"], name="e1")
    D.add_port(port=finish_se.ports["e1"], name="e2")

    D.info["num_squares"] = num_meanders * (xsize / wire_width)
    D.info["area"] = xsize * ysize
    D.info["xsize"] = xsize
    D.info["ysize"] = ysize
    D.flatten()
    return D



from __future__ import annotations

from functools import partial

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.typings import AngleInDegrees, ComponentSpec, CrossSectionSpec, Float2


@gf.cell_with_module_name
def ring_double_heater(
    gap: float = 0.2,
    gap_top: float | None = None,
    gap_bot: float | None = None,
    radius: float | None = None,
    length_x: float = 1.0,
    length_y: float = 0.01,
    coupler_ring: ComponentSpec = "coupler_ring",
    coupler_ring_top: ComponentSpec | None = None,
    straight: ComponentSpec = "straight",
    bend: ComponentSpec = "bend_euler",
    cross_section_heater: CrossSectionSpec = "heater_metal",
    cross_section_waveguide_heater: CrossSectionSpec = "strip_heater_metal",
    cross_section: CrossSectionSpec = "strip",
    via_stack: ComponentSpec = "via_stack_heater_mtop_mini",
    port_orientation: AngleInDegrees | None = None,
    via_stack_offset: Float2 = (1, 0),
    via_stack_size: Float2 | None = None,
    with_drop: bool = True,
    length_extension: float | None = None,
) -> Component:
    """Returns a double bus ring with heater on top.

    two couplers (ct: top, cb: bottom)
    connected with two vertical straights (sl: left, sr: right)

    Args:
        gap: gap between for coupler.
        gap_top: gap for the top coupler. Defaults to gap.
        gap_bot: gap for the bottom coupler. Defaults to gap.
        radius: for the bend and coupler.
        length_x: ring coupler length.
        length_y: vertical straight length.
        coupler_ring: ring coupler spec.
        coupler_ring_top: ring coupler spec for coupler away from vias (defaults to coupler_ring)
        straight: straight spec.
        bend: bend spec.
        cross_section_heater: for heater.
        cross_section_waveguide_heater: for waveguide with heater.
        cross_section: for regular waveguide.
        via_stack: for heater to routing metal.
        port_orientation: for electrical ports to promote from via_stack.
        via_stack_size: size of via_stack.
        via_stack_offset: x,y offset for via_stack.
        with_drop: adds drop ports.
        length_extension: straight length extension at the end of the coupler bottom ports.

    .. code::

           o2──────▲─────────o3
                   │gap_top
           xx──────▼─────────xxx
          xxx                   xxx
        xxx                       xxx
       xx                           xxx
       x                             xxx
      xx                              xx▲
      xx                              xx│length_y
      xx                              xx▼
      xx                             xx
       xx          length_x          x
        xx     ◄───────────────►    x
         xx                       xxx
           xx                   xxx
            xxx──────▲─────────xxx
                     │gap
             o1──────▼─────────o4
    """
    gap_top = gap_top or gap
    gap_bot = gap_bot or gap

    gap = gf.snap.snap_to_grid(gap, grid_factor=2)
    gap_top = gf.snap.snap_to_grid(gap_top, grid_factor=2)
    gap_bot = gf.snap.snap_to_grid(gap_bot, grid_factor=2)

    coupler_ring_top = coupler_ring_top or coupler_ring

    coupler_component = gf.get_component(
        coupler_ring,
        gap=gap_bot,
        radius=radius,
        length_x=length_x,
        bend=bend,
        cross_section=cross_section,
        cross_section_bend=cross_section_waveguide_heater,
        length_extension=length_extension,
    )
    coupler_component_top = gf.get_component(
        coupler_ring_top,
        gap=gap_top,
        radius=radius,
        length_x=length_x,
        bend=bend,
        cross_section=cross_section,
        cross_section_bend=cross_section_waveguide_heater,
        length_extension=length_extension,
    )
    straight_component = gf.get_component(
        straight,
        length=length_y,
        cross_section=cross_section_waveguide_heater,
    )

    c = Component()

    cb = c.add_ref(coupler_component)
    sl = c.add_ref(straight_component)
    sr = c.add_ref(straight_component)
    c.add_port("o1", port=cb.ports["o1"])
    c.add_port("o2", port=cb.ports["o4"])

    if with_drop:
        ct = c.add_ref(coupler_component_top)
        sl.connect(port="o1", other=cb.ports["o2"])
        ct.connect(port="o3", other=sl.ports["o2"])
        sr.connect(port="o2", other=ct.ports["o2"])
        c.add_port("o3", port=ct.ports["o4"])
        c.add_port("o4", port=ct.ports["o1"])
        heater_top = c << gf.get_component(
            straight,
            length=length_x,
            cross_section=cross_section_heater,
        )
        heater_top.connect("e1", ct["e1"])

    else:
        straight_top = gf.get_component(
            straight,
            length=length_x,
            cross_section=cross_section_waveguide_heater,
        )
        bend = gf.get_component(
            bend,
            radius=radius,
            cross_section=cross_section_waveguide_heater,
        )
        bl = c << bend
        br = c << bend
        st = c << straight_top

        sl.connect(port="o1", other=cb.ports["o2"])
        bl.connect(port="o2", other=sl.ports["o2"])

        st.connect(port="o2", other=bl.ports["o1"])
        br.connect(port="o2", other=st.ports["o1"])
        sr.connect(port="o1", other=br.ports["o1"])
        sr.connect(port="o2", other=cb.ports["o3"])

    if via_stack_size:
        via = gf.get_component(via_stack, size=via_stack_size)

    else:
        via = gf.get_component(via_stack)

    c1 = c << via
    c2 = c << via
    c1.xmax = -length_x / 2 + cb.x - via_stack_offset[0]
    c2.xmin = +length_x / 2 + cb.x + via_stack_offset[0]
    c1.movey(via_stack_offset[1])
    c2.movey(via_stack_offset[1])

    p1 = c1.ports.filter(orientation=port_orientation)
    p2 = c2.ports.filter(orientation=port_orientation)
    valid_orientations = {p.orientation for p in via.ports}

    if not p1:
        raise ValueError(
            f"No ports found for port_orientation {port_orientation} in {valid_orientations}"
        )

    c.flatten()
    c.add_ports(p1, prefix="l_")
    c.add_ports(p2, prefix="r_")
    return c


ring_single_heater = partial(ring_double_heater, with_drop=False)



"""CD SEM structures."""

from __future__ import annotations

from collections.abc import Sequence

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.typings import ComponentSpec, CrossSectionSpec

LINE_LENGTH = 420.0


@gf.cell_with_module_name
def cdsem_straight(
    widths: Sequence[float] = (0.4, 0.45, 0.5, 0.6, 0.8, 1.0),
    length: float = LINE_LENGTH,
    cross_section: CrossSectionSpec = "strip_no_ports",
    text: ComponentSpec | None = "text_rectangular",
    spacing: float = 7.0,
    positions: Sequence[float | None] | None = None,
    text_size: float = 1,
) -> Component:
    """Returns straight waveguide lines width sweep.

    Args:
        widths: for the sweep.
        length: for the line.
        cross_section: for the lines.
        text: optional text for labels.
        spacing: Optional center to center spacing.
        positions: Optional positions for the text labels.
        text_size: in um.
    """
    c = Component()
    p = 0.0
    if positions is not None:
        positions = positions or [None] * len(widths)
    else:
        positions = [i * spacing for i in range(len(widths))]

    for width, position in zip(widths, positions, strict=False):
        line = c << gf.c.straight(
            length=length, cross_section=cross_section, width=width
        )
        p = position or p
        line.ymin = p
        if text:
            t = c << gf.get_component(text, text=str(int(width * 1e3)), size=text_size)
            t.xmin = line.xmax + 5
            t.ymin = p

    return c



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.typings import ComponentSpec, Strs


@gf.cell(set_name=False)
def extend_ports_list(
    component_spec: ComponentSpec,
    extension: ComponentSpec,
    extension_port_name: str | None = None,
    ignore_ports: Strs | None = None,
) -> Component:
    """Returns a component with an extension attached to a list of ports.

    Args:
        component_spec: component from which to get ports.
        extension: function for extension.
        extension_port_name: to connect extension.
        ignore_ports: list of port names to ignore.
    """
    from gdsfactory.pdk import get_component

    ports = get_component(component_spec).ports

    c = Component()
    extension = get_component(extension)
    c.name = f"{extension.name}_extended_{c.cell_index()}"

    extension_port_name_or_port = extension_port_name or extension.ports[0]
    ignore_ports = ignore_ports or ()

    for i, port in enumerate(ports):
        extension_ref = c << extension
        extension_ref.connect(extension_port_name_or_port, port)

        for port in extension_ref.ports:
            port_name = port.name
            if port_name not in ignore_ports:
                c.add_port(f"{i}_{port_name}", port=port)

    c.auto_rename_ports()
    return c



"""This module contains functions to automatically add tapers to a component's ports, and to create tapers between different cross sections."""

from __future__ import annotations

import warnings
from typing import cast

import gdsfactory as gf
from gdsfactory import LayerEnum
from gdsfactory.component import Component
from gdsfactory.port import Port
from gdsfactory.typings import CrossSectionSpec, LayerSpec, LayerTransitions, Ports


def _normalize_layer_transitions(
    layer_transitions: LayerTransitions,
) -> LayerTransitions:
    normalized: LayerTransitions = {}
    for key, value in layer_transitions.items():
        normalized_key: LayerSpec | tuple[LayerEnum | int, LayerEnum | int]
        if isinstance(key, tuple) and len(key) == 2:
            if all(isinstance(k, LayerEnum) for k in key):
                normalized_key = gf.get_layer(key[0]), gf.get_layer(key[1])
            else:
                normalized_key = gf.get_layer(cast(LayerSpec, key))
        else:
            normalized_key = gf.get_layer(cast(LayerSpec, key))
        normalized[normalized_key] = value
    return normalized


def add_auto_tapers(
    component: Component,
    ports: Ports,
    cross_section: CrossSectionSpec,
    layer_transitions: LayerTransitions | None = None,
) -> list[Port]:
    """Adds tapers to the ports of a component (to be used for routing) and returns the new lists of ports.

    Args:
        component: the component to add tapers to
        ports: the list of ports
        cross_section: the cross section to route to
        layer_transitions: the layer transitions dictionary to use (use the pdk default if None)

    Returns:
        The new list of ports, on the opposite end of the tapers
    """
    cross_section = gf.get_cross_section(cross_section)
    cs_layer = gf.get_layer(cross_section.layer)
    cs_width = cross_section.width

    if layer_transitions is None:
        pdk = gf.get_active_pdk()
        layer_transitions = pdk.layer_transitions
    layer_transitions = _normalize_layer_transitions(layer_transitions)

    result: list[Port] = []

    for p in ports:
        port_layer = gf.get_layer(p.layer)
        port_width = p.width
        reverse = False

        if port_layer != cs_layer:
            try:
                taper = layer_transitions.get((port_layer, cs_layer))
                if taper is None:
                    taper = layer_transitions[cs_layer, port_layer]
                    reverse = True
            except KeyError as e:
                raise KeyError(
                    f"No registered tapers between routing layers {gf.get_layer_name(port_layer)!r} and {gf.get_layer_name(cs_layer)!r}!"
                ) from e
        elif port_width != cs_width:
            try:
                taper = layer_transitions[port_layer]
            except KeyError:
                warnings.warn(
                    f"No registered width taper for layer {port_layer}. Skipping.",
                    stacklevel=4,
                )
                result.append(p)
                continue
        else:
            result.append(p)
            continue

        if reverse:
            taper_component = gf.get_component(
                taper, width2=port_width, width1=cs_width
            )
        else:
            taper_component = gf.get_component(
                taper, width1=port_width, width2=cs_width
            )

        # Ensure we filter out i.e. extra electrical ports if we should be looking at optical ones
        taper_ports = [
            tp for tp in taper_component.ports if tp.port_type == p.port_type
        ]

        if len(taper_ports) != 2:
            raise ValueError(
                f"Taper component should have two ports of port_type={p.port_type!r}! Got {taper_component.ports}."
            )
        if taper_ports[0].layer == p.layer and taper_ports[0].width == p.width:
            p0, p1 = taper_ports
        elif taper_ports[1].layer == p.layer and taper_ports[1].width == p.width:
            p1, p0 = taper_ports
        else:
            width = p.width
            layer = p.layer
            raise ValueError(
                f"Taper component ports do not match the port's layer and width!\nTaper name: {taper_component.name}\nPorts: {taper_ports}\nCross-section: {layer=}, {width=}"
            )
        taper_ref = component.add_ref(taper_component)
        assert p0.name is not None
        taper_ref.connect(p0.name, p)
        port_new = taper_ref.ports[p1.name].copy()
        port_new.name = p.name
        result.append(port_new)
    return result


def auto_taper_to_cross_section(
    component: gf.Component,
    port: Port,
    cross_section: CrossSectionSpec,
    layer_transitions: LayerTransitions | None = None,
) -> Port:
    """Creates a taper from a port to a given cross section and places it in the component. The opposite port of the taper will be returned.

    Args:
        component: the component to place into
        port: a port to connect to, usually from a ComponentReference
        cross_section: a cross section to transition to
        layer_transitions: the layer transitions dictionary to use (use the pdk default if None)

    Returns:
        The port at the opposite (unconnected end) of the taper.
    """
    # Reuse fast version above for legacy compatibility
    # (not used in new add_auto_tapers flow)
    cross_section = gf.get_cross_section(cross_section)
    cs_layer = gf.get_layer(cross_section.layer)
    cs_width = cross_section.width

    if layer_transitions is None:
        pdk = gf.get_active_pdk()
        layer_transitions = pdk.layer_transitions
    layer_transitions = _normalize_layer_transitions(layer_transitions)

    port_layer = gf.get_layer(port.layer)
    port_width = port.width
    reverse = False

    if port_layer != cs_layer:
        try:
            taper = layer_transitions.get((port_layer, cs_layer))
            if taper is None:
                taper = layer_transitions[cs_layer, port_layer]
                reverse = True
        except KeyError as e:
            raise KeyError(
                f"No registered tapers between routing layers {gf.get_layer_name(port_layer)!r} and {gf.get_layer_name(cs_layer)!r}!"
            ) from e
    elif port_width != cs_width:
        try:
            taper = layer_transitions[port_layer]
        except KeyError:
            warnings.warn(
                f"No registered width taper for layer {port_layer}. Skipping.",
                stacklevel=4,
            )
            return port
    else:
        return port
    if reverse:
        taper_component = gf.get_component(taper, width2=port_width, width1=cs_width)
    else:
        taper_component = gf.get_component(taper, width1=port_width, width2=cs_width)

    # ensure we filter out i.e. extra electrical ports if we should be looking at optical ones
    taper_ports = [p for p in taper_component.ports if p.port_type == port.port_type]

    if len(taper_ports) != 2:
        raise ValueError(
            f"Taper component should have two ports of port_type={port.port_type!r}! Got {taper_component.ports}."
        )
    if taper_ports[0].layer == port.layer and taper_ports[0].width == port.width:
        p0, p1 = taper_ports
    elif taper_ports[1].layer == port.layer and taper_ports[1].width == port.width:
        p1, p0 = taper_ports
    else:
        width = port.width
        layer = port.layer
        raise ValueError(
            f"Taper component ports do not match the port's layer and width!\nTaper name: {taper_component.name}\nPorts: {taper_ports}\nCross-section: {layer=}, {width=}"
        )
    taper_ref = component.add_ref(taper_component)
    assert p0.name is not None
    taper_ref.connect(p0.name, port)
    port_new = taper_ref.ports[p1.name].copy()
    port_new.name = port.name
    return port_new



from gdsfactory.components import (
    bends,
    containers,
    couplers,
    dies,
    filters,
    grating_couplers,
    mmis,
    mzis,
    pads,
    pcms,
    rings,
    shapes,
    spirals,
    superconductors,
    tapers,
    texts,
    vias,
    waveguides,
)
from gdsfactory.components.analog import interdigital_capacitor
from gdsfactory.components.bends import (
    bend_circular,
    bend_circular180,
    bend_circular_all_angle,
    bend_circular_heater,
    bend_euler,
    bend_euler180,
    bend_euler_all_angle,
    bend_euler_s,
    bend_s,
    bend_s_offset,
    bezier,
)
from gdsfactory.components.containers import (
    add_fiber_array_optical_south_electrical_north,
    add_termination,
    add_trenches,
    add_trenches90,
    array,
    array_component,
    component_sequence,
    copy_layers,
    extend_ports,
    extend_ports_list,
    extension,
    generate_doe,
    line,
    pack_doe,
    pack_doe_grid,
    parse_component_name,
    splitter_chain,
    splitter_tree,
    switch_tree,
)
from gdsfactory.components.couplers import (
    coupler,
    coupler90,
    coupler90bend,
    coupler90circular,
    coupler_adiabatic,
    coupler_asymmetric,
    coupler_bent,
    coupler_broadband,
    coupler_full,
    coupler_ring,
    coupler_straight,
    coupler_straight_asymmetric,
    coupler_symmetric,
)
from gdsfactory.components.detectors import ge_detector_straight_si_contacts
from gdsfactory.components.dies import (
    add_frame,
    align,
    align_wafer,
    die,
    die_frame,
    die_frame_phix,
    die_frame_phix_dc,
    die_frame_phix_rf,
    die_frame_rf,
    die_frame_with_pads,
    die_with_pads,
    seal_ring,
    seal_ring_segmented,
    wafer,
)
from gdsfactory.components.edge_couplers import (
    edge_coupler_array,
    edge_coupler_array_with_loopback,
    edge_coupler_silicon,
)
from gdsfactory.components.filters import (
    awg,
    dbr,
    dbr_cell,
    dbr_tapered,
    fiber,
    fiber_array,
    free_propagation_region,
    loop_mirror,
    mode_converter,
    polarization_splitter_rotator,
    terminator,
    terminator_spiral,
)
from gdsfactory.components.grating_couplers import (
    grating_coupler_array,
    grating_coupler_dual_pol,
    grating_coupler_elliptical,
    grating_coupler_elliptical_arbitrary,
    grating_coupler_elliptical_lumerical,
    grating_coupler_elliptical_lumerical_etch70,
    grating_coupler_elliptical_te,
    grating_coupler_elliptical_tm,
    grating_coupler_elliptical_trenches,
    grating_coupler_elliptical_uniform,
    grating_coupler_loss,
    grating_coupler_rectangular,
    grating_coupler_rectangular_arbitrary,
    grating_coupler_te,
    grating_coupler_tm,
    grating_coupler_tree,
)
from gdsfactory.components.mmis import (
    mmi,
    mmi1x2,
    mmi1x2_with_sbend,
    mmi2x2,
    mmi2x2_with_sbend,
    mmi_90degree_hybrid,
    mmi_tapered,
    mmi_widths,
)
from gdsfactory.components.mzis import (
    mzi,
    mzi1x2,
    mzi1x2_2x2,
    mzi2x2_2x2,
    mzi2x2_2x2_phase_shifter,
    mzi_coupler,
    mzi_lattice,
    mzi_lattice_mmi,
    mzi_pads_center,
    mzi_phase_shifter,
    mzi_phase_shifter_top_heater_metal,
    mzi_pin,
    mzit,
    mzit_lattice,
    mzm,
)
from gdsfactory.components.pads import (
    pad,
    pad_array,
    pad_array0,
    pad_array90,
    pad_array180,
    pad_array270,
    pad_gs,
    pad_gsg,
    pad_gsg_open,
    pad_gsg_short,
    pad_rectangular,
    pad_small,
    pads_shorted,
    rectangle_with_slits,
)
from gdsfactory.components.pcms import (
    cavity,
    cdsem_all,
    cdsem_bend180,
    cdsem_coupler,
    cdsem_straight,
    cdsem_straight_density,
    cutback_2x2,
    cutback_bend,
    cutback_bend90,
    cutback_bend90circular,
    cutback_bend180,
    cutback_bend180circular,
    cutback_component,
    cutback_component_mirror,
    cutback_loss,
    cutback_loss_bend90,
    cutback_loss_bend180,
    cutback_loss_mmi1x2,
    cutback_loss_spirals,
    cutback_splitter,
    greek_cross,
    greek_cross_with_pads,
    litho_calipers,
    litho_ruler,
    litho_steps,
    pixel,
    qrcode,
    resistance_meander,
    resistance_sheet,
    ruler,
    staircase,
    verniers,
    version_stamp,
    widths,
)
from gdsfactory.components.quantum import (
    coupler_capacitive,
    coupler_interdigital,
    coupler_tunable,
    flux_qubit,
    flux_qubit_asymmetric,
    resonator_cpw,
    resonator_lumped,
    resonator_quarter_wave,
    transmon,
    transmon_circular,
)
from gdsfactory.components.rings import (
    coupler_bend,
    coupler_ring_bend,
    disk,
    disk_heater,
    ring,
    ring_asymmetric,
    ring_crow,
    ring_crow_couplers,
    ring_double,
    ring_double_bend_coupler,
    ring_double_heater,
    ring_double_pn,
    ring_heater,
    ring_single,
    ring_single_array,
    ring_single_bend_coupler,
    ring_single_dut,
    ring_single_heater,
    ring_single_pn,
)
from gdsfactory.components.shapes import (
    C,
    L,
    bbox,
    bbox_to_points,
    circle,
    compass,
    cross,
    ellipse,
    fiber_size,
    fiducial_squares,
    hexagon,
    marker_te,
    marker_tm,
    nxn,
    octagon,
    rectangle,
    rectangles,
    regular_polygon,
    triangle,
    triangle2,
    triangle2_thin,
    triangle4,
    triangle4_thin,
    triangle_thin,
    triangles,
)
from gdsfactory.components.spirals import (
    delay_snake,
    delay_snake2,
    delay_snake_sbend,
    spiral,
    spiral_double,
    spiral_heater,
    spiral_inductor,
    spiral_racetrack,
    spiral_racetrack_fixed_length,
    spiral_racetrack_heater_doped,
    spiral_racetrack_heater_metal,
)
from gdsfactory.components.superconductors import (
    hline,
    optimal_90deg,
    optimal_hairpin,
    optimal_step,
    snspd,
)
from gdsfactory.components.tapers import (
    ramp,
    taper,
    taper_0p5_to_3_l36,
    taper_adiabatic,
    taper_cross_section,
    taper_cross_section_linear,
    taper_cross_section_parabolic,
    taper_cross_section_sine,
    taper_electrical,
    taper_from_csv,
    taper_nc_sc,
    taper_parabolic,
    taper_sc_nc,
    taper_strip_to_ridge,
    taper_strip_to_ridge_trenches,
    taper_strip_to_slab150,
    taper_w10_l100,
    taper_w10_l150,
    taper_w10_l200,
    taper_w11_l200,
    taper_w12_l200,
)
from gdsfactory.components.texts import (
    character_a,
    pixel_array,
    rectangular_font,
    text,
    text_freetype,
    text_klayout,
    text_lines,
    text_rectangular,
    text_rectangular_font,
    text_rectangular_multi_layer,
)
from gdsfactory.components.vias import (
    via,
    via1,
    via2,
    via_chain,
    via_circular,
    via_corner,
    via_stack,
    via_stack_corner45,
    via_stack_corner45_extended,
    via_stack_heater_m2,
    via_stack_heater_m3,
    via_stack_heater_mtop,
    via_stack_heater_mtop_mini,
    via_stack_m1_m3,
    via_stack_m1_mtop,
    via_stack_m2_m3,
    via_stack_npp_m1,
    via_stack_slab_m1,
    via_stack_slab_m1_horizontal,
    via_stack_slab_m2,
    via_stack_slab_m3,
    via_stack_slab_npp_m3,
    via_stack_with_offset,
    via_stack_with_offset_m1_m3,
    via_stack_with_offset_ppp_m1,
    viac,
)
from gdsfactory.components.waveguides import (
    crossing,
    crossing45,
    crossing_etched,
    crossing_linear_taper,
    crossing_waveguide,
    straight,
    straight_all_angle,
    straight_array,
    straight_heater_doped,
    straight_heater_doped_rib,
    straight_heater_doped_strip,
    straight_heater_meander,
    straight_heater_meander_doped,
    straight_heater_metal,
    straight_heater_metal_90_90,
    straight_heater_metal_simple,
    straight_heater_metal_undercut,
    straight_heater_metal_undercut_90_90,
    straight_piecewise,
    straight_pin,
    straight_pin_slot,
    straight_pn,
    straight_pn_slot,
    wire,
    wire_corner,
    wire_corner45,
    wire_corner45_straight,
    wire_corner_sections,
    wire_straight,
)

__all__ = [
    "C",
    "L",
    "add_fiber_array_optical_south_electrical_north",
    "add_frame",
    "add_termination",
    "add_trenches",
    "add_trenches90",
    "align",
    "align_wafer",
    "array",
    "array_component",
    "awg",
    "bbox",
    "bbox_to_points",
    "bend_circular",
    "bend_circular180",
    "bend_circular_all_angle",
    "bend_circular_heater",
    "bend_euler",
    "bend_euler180",
    "bend_euler_all_angle",
    "bend_euler_s",
    "bend_s",
    "bend_s_offset",
    "bends",
    "bezier",
    "cavity",
    "cdsem_all",
    "cdsem_bend180",
    "cdsem_coupler",
    "cdsem_straight",
    "cdsem_straight_density",
    "character_a",
    "circle",
    "compass",
    "component_sequence",
    "containers",
    "copy_layers",
    "coupler",
    "coupler90",
    "coupler90bend",
    "coupler90circular",
    "coupler_adiabatic",
    "coupler_asymmetric",
    "coupler_bend",
    "coupler_bent",
    "coupler_broadband",
    "coupler_capacitive",
    "coupler_full",
    "coupler_interdigital",
    "coupler_ring",
    "coupler_ring_bend",
    "coupler_straight",
    "coupler_straight_asymmetric",
    "coupler_symmetric",
    "coupler_tunable",
    "couplers",
    "cross",
    "crossing",
    "crossing45",
    "crossing_etched",
    "crossing_linear_taper",
    "crossing_waveguide",
    "cutback_2x2",
    "cutback_bend",
    "cutback_bend90",
    "cutback_bend90circular",
    "cutback_bend180",
    "cutback_bend180circular",
    "cutback_component",
    "cutback_component_mirror",
    "cutback_loss",
    "cutback_loss_bend90",
    "cutback_loss_bend180",
    "cutback_loss_mmi1x2",
    "cutback_loss_spirals",
    "cutback_splitter",
    "dbr",
    "dbr_cell",
    "dbr_tapered",
    "delay_snake",
    "delay_snake2",
    "delay_snake_sbend",
    "die",
    "die_frame",
    "die_frame_phix",
    "die_frame_phix_dc",
    "die_frame_phix_rf",
    "die_frame_rf",
    "die_frame_with_pads",
    "die_with_pads",
    "dies",
    "disk",
    "disk_heater",
    "edge_coupler_array",
    "edge_coupler_array_with_loopback",
    "edge_coupler_silicon",
    "ellipse",
    "extend_ports",
    "extend_ports_list",
    "extension",
    "fiber",
    "fiber_array",
    "fiber_size",
    "fiducial_squares",
    "filters",
    "flux_qubit",
    "flux_qubit_asymmetric",
    "free_propagation_region",
    "ge_detector_straight_si_contacts",
    "generate_doe",
    "grating_coupler_array",
    "grating_coupler_dual_pol",
    "grating_coupler_elliptical",
    "grating_coupler_elliptical_arbitrary",
    "grating_coupler_elliptical_lumerical",
    "grating_coupler_elliptical_lumerical_etch70",
    "grating_coupler_elliptical_te",
    "grating_coupler_elliptical_tm",
    "grating_coupler_elliptical_trenches",
    "grating_coupler_elliptical_uniform",
    "grating_coupler_loss",
    "grating_coupler_rectangular",
    "grating_coupler_rectangular_arbitrary",
    "grating_coupler_te",
    "grating_coupler_tm",
    "grating_coupler_tree",
    "grating_couplers",
    "greek_cross",
    "greek_cross_with_pads",
    "hexagon",
    "hline",
    "interdigital_capacitor",
    "line",
    "litho_calipers",
    "litho_ruler",
    "litho_steps",
    "loop_mirror",
    "marker_te",
    "marker_tm",
    "mmi",
    "mmi1x2",
    "mmi1x2_with_sbend",
    "mmi2x2",
    "mmi2x2_with_sbend",
    "mmi_90degree_hybrid",
    "mmi_tapered",
    "mmi_widths",
    "mmis",
    "mode_converter",
    "mzi",
    "mzi1x2",
    "mzi1x2_2x2",
    "mzi2x2_2x2",
    "mzi2x2_2x2_phase_shifter",
    "mzi_coupler",
    "mzi_lattice",
    "mzi_lattice_mmi",
    "mzi_pads_center",
    "mzi_phase_shifter",
    "mzi_phase_shifter_top_heater_metal",
    "mzi_pin",
    "mzis",
    "mzit",
    "mzit_lattice",
    "mzm",
    "nxn",
    "octagon",
    "optimal_90deg",
    "optimal_hairpin",
    "optimal_step",
    "pack_doe",
    "pack_doe_grid",
    "pad",
    "pad_array",
    "pad_array0",
    "pad_array90",
    "pad_array180",
    "pad_array270",
    "pad_gs",
    "pad_gsg",
    "pad_gsg_open",
    "pad_gsg_short",
    "pad_rectangular",
    "pad_small",
    "pads",
    "pads_shorted",
    "parse_component_name",
    "pcms",
    "pixel",
    "pixel_array",
    "polarization_splitter_rotator",
    "qrcode",
    "ramp",
    "rectangle",
    "rectangle_with_slits",
    "rectangles",
    "rectangular_font",
    "regular_polygon",
    "resistance_meander",
    "resistance_sheet",
    "resonator_cpw",
    "resonator_lumped",
    "resonator_quarter_wave",
    "ring",
    "ring_asymmetric",
    "ring_crow",
    "ring_crow_couplers",
    "ring_double",
    "ring_double_bend_coupler",
    "ring_double_heater",
    "ring_double_pn",
    "ring_heater",
    "ring_single",
    "ring_single_array",
    "ring_single_bend_coupler",
    "ring_single_dut",
    "ring_single_heater",
    "ring_single_pn",
    "rings",
    "ruler",
    "seal_ring",
    "seal_ring_segmented",
    "shapes",
    "snspd",
    "spiral",
    "spiral_double",
    "spiral_heater",
    "spiral_inductor",
    "spiral_racetrack",
    "spiral_racetrack_fixed_length",
    "spiral_racetrack_heater_doped",
    "spiral_racetrack_heater_metal",
    "spirals",
    "splitter_chain",
    "splitter_tree",
    "staircase",
    "straight",
    "straight_all_angle",
    "straight_array",
    "straight_heater_doped",
    "straight_heater_doped_rib",
    "straight_heater_doped_strip",
    "straight_heater_meander",
    "straight_heater_meander_doped",
    "straight_heater_metal",
    "straight_heater_metal_90_90",
    "straight_heater_metal_simple",
    "straight_heater_metal_undercut",
    "straight_heater_metal_undercut_90_90",
    "straight_piecewise",
    "straight_pin",
    "straight_pin_slot",
    "straight_pn",
    "straight_pn_slot",
    "superconductors",
    "switch_tree",
    "taper",
    "taper_0p5_to_3_l36",
    "taper_adiabatic",
    "taper_cross_section",
    "taper_cross_section_linear",
    "taper_cross_section_parabolic",
    "taper_cross_section_sine",
    "taper_electrical",
    "taper_from_csv",
    "taper_nc_sc",
    "taper_parabolic",
    "taper_sc_nc",
    "taper_strip_to_ridge",
    "taper_strip_to_ridge_trenches",
    "taper_strip_to_slab150",
    "taper_w10_l100",
    "taper_w10_l150",
    "taper_w10_l200",
    "taper_w11_l200",
    "taper_w12_l200",
    "tapers",
    "terminator",
    "terminator_spiral",
    "text",
    "text_freetype",
    "text_klayout",
    "text_lines",
    "text_rectangular",
    "text_rectangular_font",
    "text_rectangular_multi_layer",
    "texts",
    "transmon",
    "transmon_circular",
    "triangle",
    "triangle2",
    "triangle2_thin",
    "triangle4",
    "triangle4_thin",
    "triangle_thin",
    "triangles",
    "verniers",
    "version_stamp",
    "via",
    "via1",
    "via2",
    "via_chain",
    "via_circular",
    "via_corner",
    "via_stack",
    "via_stack_corner45",
    "via_stack_corner45_extended",
    "via_stack_heater_m2",
    "via_stack_heater_m3",
    "via_stack_heater_mtop",
    "via_stack_heater_mtop_mini",
    "via_stack_m1_m3",
    "via_stack_m1_mtop",
    "via_stack_m2_m3",
    "via_stack_npp_m1",
    "via_stack_slab_m1",
    "via_stack_slab_m1_horizontal",
    "via_stack_slab_m2",
    "via_stack_slab_m3",
    "via_stack_slab_npp_m3",
    "via_stack_with_offset",
    "via_stack_with_offset_m1_m3",
    "via_stack_with_offset_ppp_m1",
    "viac",
    "vias",
    "wafer",
    "waveguides",
    "widths",
    "wire",
    "wire_corner",
    "wire_corner45",
    "wire_corner45_straight",
    "wire_corner_sections",
    "wire_straight",
]
